# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.dep import Atom, ExtendedAtomDict, best_match_to_list, match_from_list
from portage.exception import InvalidAtom
from portage.versions import cpv_getkey


OPERATIONS = ["merge", "unmerge"]


class PackageSet:
    # Set this to operations that are supported by your subclass. While
    # technically there is no difference between "merge" and "unmerge" regarding
    # package sets, the latter doesn't make sense for some sets like "system"
    # or "security" and therefore isn't supported by them.
    _operations = ["merge"]
    description = "generic package set"

    def __init__(self, allow_wildcard=False, allow_repo=False):
        self._atoms = set()
        self._atommap = ExtendedAtomDict(set)
        self._loaded = False
        self._loading = False
        self.errors = []
        self._nonatoms = set()
        self.world_candidate = False
        self._allow_wildcard = allow_wildcard
        self._allow_repo = allow_repo

    def __contains__(self, atom):
        self._load()
        return atom in self._atoms or atom in self._nonatoms

    def __iter__(self):
        self._load()
        for x in self._atoms:
            yield x
        for x in self._nonatoms:
            yield x

    def __bool__(self):
        self._load()
        return bool(self._atoms or self._nonatoms)

    def supportsOperation(self, op):
        if not op in OPERATIONS:
            raise ValueError(op)
        return op in self._operations

    def _load(self):
        if not (self._loaded or self._loading):
            self._loading = True
            self.load()
            self._loaded = True
            self._loading = False

    def getAtoms(self):
        self._load()
        return self._atoms.copy()

    def getNonAtoms(self):
        self._load()
        return self._nonatoms.copy()

    def _setAtoms(self, atoms):
        self._atoms.clear()
        self._nonatoms.clear()
        for a in atoms:
            if not isinstance(a, Atom):
                if isinstance(a, str):
                    a = a.strip()
                if not a:
                    continue
                try:
                    a = Atom(a, allow_wildcard=True, allow_repo=True)
                except InvalidAtom:
                    self._nonatoms.add(a)
                    continue
            if not self._allow_wildcard and a.extended_syntax:
                raise InvalidAtom("extended atom syntax not allowed here")
            if not self._allow_repo and a.repo:
                raise InvalidAtom("repository specification not allowed here")
            self._atoms.add(a)

        self._updateAtomMap()

    def load(self):
        # This method must be overwritten by subclasses
        # Editable sets should use the value of self._mtime to determine if they
        # need to reload themselves
        raise NotImplementedError()

    def containsCPV(self, cpv):
        self._load()
        for a in self._atoms:
            if match_from_list(a, [cpv]):
                return True
        return False

    def getMetadata(self, key):
        if hasattr(self, key.lower()):
            return getattr(self, key.lower())
        return ""

    def _updateAtomMap(self, atoms=None):
        """Update self._atommap for specific atoms or all atoms."""
        if not atoms:
            self._atommap.clear()
            atoms = self._atoms
        for a in atoms:
            self._atommap.setdefault(a.cp, set()).add(a)

    # Not sure if this one should really be in PackageSet
    def findAtomForPackage(self, pkg, modified_use=None):
        """Return the best match for a given package from the arguments, or
        None if there are no matches.  This matches virtual arguments against
        the PROVIDE metadata.  This can raise an InvalidDependString exception
        if an error occurs while parsing PROVIDE."""

        if modified_use is not None and modified_use is not pkg.use.enabled:
            pkg = pkg.copy()
            pkg._metadata["USE"] = " ".join(modified_use)

        # Atoms matched via PROVIDE must be temporarily transformed since
        # match_from_list() only works correctly when atom.cp == pkg.cp.
        rev_transform = {}
        for atom in self.iterAtomsForPackage(pkg):
            if atom.cp == pkg.cp:
                rev_transform[atom] = atom
            else:
                rev_transform[
                    Atom(
                        atom.replace(atom.cp, pkg.cp, 1),
                        allow_wildcard=True,
                        allow_repo=True,
                    )
                ] = atom
        best_match = best_match_to_list(pkg, iter(rev_transform))
        if best_match:
            return rev_transform[best_match]
        return None

    def iterAtomsForPackage(self, pkg):
        """
        Find all matching atoms for a given package. This matches virtual
        arguments against the PROVIDE metadata.  This will raise an
        InvalidDependString exception if PROVIDE is invalid.
        """
        cpv_slot_list = [pkg]
        cp = cpv_getkey(pkg.cpv)
        self._load()  # make sure the atoms are loaded

        atoms = self._atommap.get(cp)
        if atoms:
            for atom in atoms:
                if match_from_list(atom, cpv_slot_list):
                    yield atom


class EditablePackageSet(PackageSet):
    def __init__(self, allow_wildcard=False, allow_repo=False):
        super(EditablePackageSet, self).__init__(
            allow_wildcard=allow_wildcard, allow_repo=allow_repo
        )

    def update(self, atoms):
        self._load()
        modified = False
        normal_atoms = []
        for a in atoms:
            if not isinstance(a, Atom):
                try:
                    a = Atom(a, allow_wildcard=True, allow_repo=True)
                except InvalidAtom:
                    modified = True
                    self._nonatoms.add(a)
                    continue
            if not self._allow_wildcard and a.extended_syntax:
                raise InvalidAtom("extended atom syntax not allowed here")
            if not self._allow_repo and a.repo:
                raise InvalidAtom("repository specification not allowed here")
            normal_atoms.append(a)

        if normal_atoms:
            modified = True
            self._atoms.update(normal_atoms)
            self._updateAtomMap(atoms=normal_atoms)
        if modified:
            self.write()

    def add(self, atom):
        self.update([atom])

    def replace(self, atoms):
        self._setAtoms(atoms)
        self.write()

    def remove(self, atom):
        self._load()
        self._atoms.discard(atom)
        self._nonatoms.discard(atom)
        self._updateAtomMap()
        self.write()

    def removePackageAtoms(self, cp):
        self._load()
        for a in list(self._atoms):
            if a.cp == cp:
                self.remove(a)
        self.write()

    def write(self):
        # This method must be overwritten in subclasses that should be editable
        raise NotImplementedError()


class InternalPackageSet(EditablePackageSet):
    def __init__(self, initial_atoms=None, allow_wildcard=False, allow_repo=True):
        """
        Repo atoms are allowed more often than not, so it makes sense for this
        class to allow them by default. The Atom constructor and isvalidatom()
        functions default to allow_repo=False, which is sufficient to ensure
        that repo atoms are prohibited when necessary.
        """
        super(InternalPackageSet, self).__init__(
            allow_wildcard=allow_wildcard, allow_repo=allow_repo
        )
        if initial_atoms is not None:
            self.update(initial_atoms)

    def clear(self):
        self._atoms.clear()
        self._updateAtomMap()

    def load(self):
        pass

    def write(self):
        pass


class DummyPackageSet(PackageSet):
    def __init__(self, atoms=None):
        super(DummyPackageSet, self).__init__()
        if atoms:
            self._setAtoms(atoms)

    def load(self):
        pass

    def singleBuilder(cls, options, settings, trees):
        atoms = options.get("packages", "").split()
        return DummyPackageSet(atoms=atoms)

    singleBuilder = classmethod(singleBuilder)



# Copyright 2012-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("global_event_loop",)

from portage.util.futures._asyncio import _safe_loop as global_event_loop



# test_dep_getslot.py -- Portage Unit Testing Functionality
# Copyright 2006 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import dep_getslot


class DepGetSlot(TestCase):
    """A simple testcase for isvalidatom"""

    def testDepGetSlot(self):

        slot_char = ":"
        slots = ("a", "1.2", "1", "IloveVapier", None)
        cpvs = ["sys-apps/portage"]
        versions = ["2.1.1", "2.1-r1"]
        for cpv in cpvs:
            for version in versions:
                for slot in slots:
                    mycpv = cpv
                    if version:
                        mycpv = "=" + mycpv + "-" + version
                    if slot is not None:
                        self.assertEqual(dep_getslot(mycpv + slot_char + slot), slot)
                    else:
                        self.assertEqual(dep_getslot(mycpv), slot)



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import fcntl
import logging
import sys

try:
    import dummy_threading
except ImportError:
    dummy_threading = None

try:
    import threading
except ImportError:
    threading = dummy_threading

import portage
from portage import os
from portage.exception import TryAgain
from portage.localization import _
from portage.locks import lockfile, unlockfile
from portage.util import writemsg_level
from _emerge.AbstractPollTask import AbstractPollTask
from _emerge.AsynchronousTask import AsynchronousTask
from _emerge.SpawnProcess import SpawnProcess


class AsynchronousLock(AsynchronousTask):
    """
    This uses the portage.locks module to acquire a lock asynchronously,
    using either a thread (if available) or a subprocess.

    The default behavior is to use a process instead of a thread, since
    there is currently no way to interrupt a thread that is waiting for
    a lock (notably, SIGINT doesn't work because python delivers all
    signals to the main thread).
    """

    __slots__ = ("path",) + (
        "_imp",
        "_force_async",
        "_force_dummy",
        "_force_process",
        "_force_thread",
        "_unlock_future",
    )

    _use_process_by_default = True

    def _start(self):

        if not self._force_async:
            try:
                self._imp = lockfile(
                    self.path, wantnewlockfile=True, flags=os.O_NONBLOCK
                )
            except TryAgain:
                pass
            else:
                self.returncode = os.EX_OK
                self._async_wait()
                return

        if self._force_process or (
            not self._force_thread
            and (self._use_process_by_default or threading is dummy_threading)
        ):
            self._imp = _LockProcess(path=self.path, scheduler=self.scheduler)
        else:
            self._imp = _LockThread(
                path=self.path, scheduler=self.scheduler, _force_dummy=self._force_dummy
            )

        self._imp.addExitListener(self._imp_exit)
        self._imp.start()

    def _imp_exit(self, imp):
        # call exit listeners
        self.returncode = imp.returncode
        self._async_wait()

    def _cancel(self):
        if isinstance(self._imp, AsynchronousTask):
            self._imp.cancel()

    def _poll(self):
        if isinstance(self._imp, AsynchronousTask):
            self._imp.poll()
        return self.returncode

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        if self._imp is None:
            raise AssertionError("not locked")
        if self._unlock_future is not None:
            raise AssertionError("already unlocked")
        if isinstance(self._imp, (_LockProcess, _LockThread)):
            unlock_future = self._imp.async_unlock()
        else:
            unlockfile(self._imp)
            unlock_future = self.scheduler.create_future()
            self.scheduler.call_soon(unlock_future.set_result, None)
        self._imp = None
        self._unlock_future = unlock_future
        return unlock_future


class _LockThread(AbstractPollTask):
    """
    This uses the portage.locks module to acquire a lock asynchronously,
    using a background thread. After the lock is acquired, the thread
    writes to a pipe in order to notify a poll loop running in the main
    thread.

    If the threading module is unavailable then the dummy_threading
    module will be used, and the lock will be acquired synchronously
    (before the start() method returns).
    """

    __slots__ = ("path",) + ("_force_dummy", "_lock_obj", "_thread", "_unlock_future")

    def _start(self):
        self._registered = True
        threading_mod = threading
        if self._force_dummy:
            threading_mod = dummy_threading
        self._thread = threading_mod.Thread(target=self._run_lock)
        self._thread.daemon = True
        self._thread.start()

    def _run_lock(self):
        self._lock_obj = lockfile(self.path, wantnewlockfile=True)
        # Thread-safe callback to EventLoop
        self.scheduler.call_soon_threadsafe(self._run_lock_cb)

    def _run_lock_cb(self):
        self._unregister()
        self.returncode = os.EX_OK
        self._async_wait()

    def _cancel(self):
        # There's currently no way to force thread termination.
        pass

    def _unlock(self):
        if self._lock_obj is None:
            raise AssertionError("not locked")
        if self.returncode is None:
            raise AssertionError("lock not acquired yet")
        if self._unlock_future is not None:
            raise AssertionError("already unlocked")
        self._unlock_future = self.scheduler.create_future()
        unlockfile(self._lock_obj)
        self._lock_obj = None

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        self._unlock()
        self.scheduler.call_soon(self._unlock_future.set_result, None)
        return self._unlock_future

    def _unregister(self):
        self._registered = False

        if self._thread is not None:
            self._thread.join()
            self._thread = None


class _LockProcess(AbstractPollTask):
    """
    This uses the portage.locks module to acquire a lock asynchronously,
    using a subprocess. After the lock is acquired, the process
    writes to a pipe in order to notify a poll loop running in the main
    process. The unlock() method notifies the subprocess to release the
    lock and exit.
    """

    __slots__ = ("path",) + (
        "_acquired",
        "_kill_test",
        "_proc",
        "_files",
        "_unlock_future",
    )

    def _start(self):
        in_pr, in_pw = os.pipe()
        out_pr, out_pw = os.pipe()
        self._files = {}
        self._files["pipe_in"] = in_pr
        self._files["pipe_out"] = out_pw

        fcntl.fcntl(
            in_pr, fcntl.F_SETFL, fcntl.fcntl(in_pr, fcntl.F_GETFL) | os.O_NONBLOCK
        )

        self.scheduler.add_reader(in_pr, self._output_handler)
        self._registered = True
        self._proc = SpawnProcess(
            args=[
                portage._python_interpreter,
                os.path.join(portage._bin_path, "lock-helper.py"),
                self.path,
            ],
            env=dict(os.environ, PORTAGE_PYM_PATH=portage._pym_path),
            fd_pipes={0: out_pr, 1: in_pw, 2: sys.__stderr__.fileno()},
            scheduler=self.scheduler,
        )
        self._proc.addExitListener(self._proc_exit)
        self._proc.start()
        os.close(out_pr)
        os.close(in_pw)

    def _proc_exit(self, proc):

        if self._files is not None:
            # Close pipe_out if it's still open, since it's useless
            # after the process has exited. This helps to avoid
            # "ResourceWarning: unclosed file" since Python 3.2.
            try:
                pipe_out = self._files.pop("pipe_out")
            except KeyError:
                pass
            else:
                os.close(pipe_out)

        if proc.returncode != os.EX_OK:
            # Typically, this will happen due to the
            # process being killed by a signal.

            if not self._acquired:
                # If the lock hasn't been aquired yet, the
                # caller can check the returncode and handle
                # this failure appropriately.
                if not (self.cancelled or self._kill_test):
                    writemsg_level(
                        "_LockProcess: %s\n"
                        % _("failed to acquire lock on '%s'")
                        % (self.path,),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                self._unregister()
                self.returncode = proc.returncode
                self._async_wait()
                return

            if not self.cancelled and self._unlock_future is None:
                # We don't want lost locks going unnoticed, so it's
                # only safe to ignore if either the cancel() or
                # unlock() methods have been previously called.
                raise AssertionError(
                    "lock process failed with returncode %s" % (proc.returncode,)
                )

        if self._unlock_future is not None:
            self._unlock_future.set_result(None)

    def _cancel(self):
        if self._proc is not None:
            self._proc.cancel()

    def _poll(self):
        if self._proc is not None:
            self._proc.poll()
        return self.returncode

    def _output_handler(self):
        buf = self._read_buf(self._files["pipe_in"])
        if buf:
            self._acquired = True
            self._unregister()
            self.returncode = os.EX_OK
            self._async_wait()

        return True

    def _unregister(self):
        self._registered = False

        if self._files is not None:
            try:
                pipe_in = self._files.pop("pipe_in")
            except KeyError:
                pass
            else:
                self.scheduler.remove_reader(pipe_in)
                os.close(pipe_in)

    def _unlock(self):
        if self._proc is None:
            raise AssertionError("not locked")
        if not self._acquired:
            raise AssertionError("lock not acquired yet")
        if self.returncode != os.EX_OK:
            raise AssertionError(
                "lock process failed with returncode %s" % (self.returncode,)
            )
        if self._unlock_future is not None:
            raise AssertionError("already unlocked")
        self._unlock_future = self.scheduler.create_future()
        os.write(self._files["pipe_out"], b"\0")
        os.close(self._files["pipe_out"])
        self._files = None

    def async_unlock(self):
        """
        Release the lock asynchronously. Release notification is available
        via the add_done_callback method of the returned Future instance.

        @returns: Future, result is None
        """
        self._unlock()
        return self._unlock_future



# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "case_insensitive_vars",
    "default_globals",
    "env_blacklist",
    "environ_filter",
    "environ_whitelist",
    "environ_whitelist_re",
)

import re

# Blacklisted variables are internal variables that are never allowed
# to enter the config instance from the external environment or
# configuration files.
env_blacklist = frozenset(
    (
        "A",
        "AA",
        "BASH_FUNC____in_portage_iuse%%",
        "BDEPEND",
        "BROOT",
        "CATEGORY",
        "DEPEND",
        "DESCRIPTION",
        "DOCS",
        "EAPI",
        "EBUILD_FORCE_TEST",
        "EBUILD_PHASE",
        "EBUILD_PHASE_FUNC",
        "EBUILD_SKIP_MANIFEST",
        "ED",
        "EMERGE_FROM",
        "EPREFIX",
        "EROOT",
        "GREP_OPTIONS",
        "HOMEPAGE",
        "IDEPEND",
        "INHERITED",
        "IUSE",
        "IUSE_EFFECTIVE",
        "KEYWORDS",
        "LICENSE",
        "MERGE_TYPE",
        "PDEPEND",
        "PF",
        "PKGUSE",
        "PORTAGE_BACKGROUND",
        "PORTAGE_BACKGROUND_UNMERGE",
        "PORTAGE_BUILDDIR_LOCKED",
        "PORTAGE_BUILT_USE",
        "PORTAGE_CONFIGROOT",
        "PORTAGE_EXPLICIT_INHERIT",
        "PORTAGE_INTERNAL_CALLER",
        "PORTAGE_IUSE",
        "PORTAGE_NONFATAL",
        "PORTAGE_PIPE_FD",
        "PORTAGE_REPO_NAME",
        "PORTAGE_USE",
        "PROPERTIES",
        "RDEPEND",
        "REPOSITORY",
        "REQUIRED_USE",
        "RESTRICT",
        "ROOT",
        "SANDBOX_LOG",
        "SLOT",
        "SRC_URI",
        "_",
    )
)

environ_whitelist = []

# Whitelisted variables are always allowed to enter the ebuild
# environment. Generally, this only includes special portage
# variables. Ebuilds can unset variables that are not whitelisted
# and rely on them remaining unset for future phases, without them
# leaking back in from various locations (bug #189417). It's very
# important to set our special BASH_ENV variable in the ebuild
# environment in order to prevent sandbox from sourcing /etc/profile
# in it's bashrc (causing major leakage).
environ_whitelist += [
    "ACCEPT_LICENSE",
    "BASH_ENV",
    "BASH_FUNC____in_portage_iuse%%",
    "BINPKG_FORMAT",
    "BROOT",
    "BUILD_ID",
    "BUILD_PREFIX",
    "COLUMNS",
    "D",
    "DISTDIR",
    "DOC_SYMLINKS_DIR",
    "EAPI",
    "EBUILD",
    "EBUILD_FORCE_TEST",
    "EBUILD_PHASE",
    "EBUILD_PHASE_FUNC",
    "ECLASSDIR",
    "ECLASS_DEPTH",
    "ED",
    "EMERGE_FROM",
    "ENV_UNSET",
    "EPREFIX",
    "EROOT",
    "ESYSROOT",
    "FEATURES",
    "FILESDIR",
    "HOME",
    "MERGE_TYPE",
    "NOCOLOR",
    "PATH",
    "PKGDIR",
    "PKGUSE",
    "PKG_LOGDIR",
    "PKG_TMPDIR",
    "PORTAGE_ACTUAL_DISTDIR",
    "PORTAGE_ARCHLIST",
    "PORTAGE_BASHRC_FILES",
    "PORTAGE_BASHRC",
    "PM_EBUILD_HOOK_DIR",
    "PORTAGE_BINPKG_FILE",
    "PORTAGE_BINPKG_TAR_OPTS",
    "PORTAGE_BINPKG_TMPFILE",
    "PORTAGE_BIN_PATH",
    "PORTAGE_BUILDDIR",
    "PORTAGE_BUILD_GROUP",
    "PORTAGE_BUILD_USER",
    "PORTAGE_BUNZIP2_COMMAND",
    "PORTAGE_BZIP2_COMMAND",
    "PORTAGE_COLORMAP",
    "PORTAGE_COMPRESS",
    "PORTAGE_COMPRESSION_COMMAND",
    "PORTAGE_COMPRESS_EXCLUDE_SUFFIXES",
    "PORTAGE_CONFIGROOT",
    "PORTAGE_DEBUG",
    "PORTAGE_DEPCACHEDIR",
    "PORTAGE_DOHTML_UNWARNED_SKIPPED_EXTENSIONS",
    "PORTAGE_DOHTML_UNWARNED_SKIPPED_FILES",
    "PORTAGE_DOHTML_WARN_ON_SKIPPED_FILES",
    "PORTAGE_EBUILD_EXIT_FILE",
    "PORTAGE_FEATURES",
    "PORTAGE_GID",
    "PORTAGE_GRPNAME",
    "PORTAGE_INTERNAL_CALLER",
    "PORTAGE_INST_GID",
    "PORTAGE_INST_UID",
    "PORTAGE_IPC_DAEMON",
    "PORTAGE_IUSE",
    "PORTAGE_ECLASS_LOCATIONS",
    "PORTAGE_LOG_FILE",
    "PORTAGE_OVERRIDE_EPREFIX",
    "PORTAGE_PIPE_FD",
    "PORTAGE_PROPERTIES",
    "PORTAGE_PYM_PATH",
    "PORTAGE_PYTHON",
    "PORTAGE_PYTHONPATH",
    "PORTAGE_QUIET",
    "PORTAGE_REPO_NAME",
    "PORTAGE_REPOSITORIES",
    "PORTAGE_RESTRICT",
    "PORTAGE_SIGPIPE_STATUS",
    "PORTAGE_SOCKS5_PROXY",
    "PORTAGE_TMPDIR",
    "PORTAGE_UPDATE_ENV",
    "PORTAGE_USERNAME",
    "PORTAGE_VERBOSE",
    "PORTAGE_WORKDIR_MODE",
    "PORTAGE_XATTR_EXCLUDE",
    "PORTDIR",
    "PORTDIR_OVERLAY",
    "PREROOTPATH",
    "PYTHONDONTWRITEBYTECODE",
    "REPLACING_VERSIONS",
    "REPLACED_BY_VERSION",
    "ROOT",
    "ROOTPATH",
    "SANDBOX_LOG",
    "SYSROOT",
    "T",
    "TMP",
    "TMPDIR",
    "USE_EXPAND",
    "USE_ORDER",
    "WORKDIR",
    "XARGS",
    "__PORTAGE_TEST_HARDLINK_LOCKS",
]

# user config variables
environ_whitelist += ["DOC_SYMLINKS_DIR", "INSTALL_MASK", "PKG_INSTALL_MASK"]

environ_whitelist += ["A", "AA", "CATEGORY", "P", "PF", "PN", "PR", "PV", "PVR"]

# misc variables inherited from the calling environment
environ_whitelist += [
    "COLORTERM",
    "DISPLAY",
    "EDITOR",
    "LESS",
    "LESSOPEN",
    "LOGNAME",
    "LS_COLORS",
    "PAGER",
    "TERM",
    "TERMCAP",
    "USER",
    "ftp_proxy",
    "http_proxy",
    "no_proxy",
]

# tempdir settings
environ_whitelist += [
    "TMPDIR",
    "TEMP",
    "TMP",
]

# localization settings
environ_whitelist += [
    "LANG",
    "LC_COLLATE",
    "LC_CTYPE",
    "LC_MESSAGES",
    "LC_MONETARY",
    "LC_NUMERIC",
    "LC_TIME",
    "LC_PAPER",
    "LC_ALL",
]

# other variables inherited from the calling environment
environ_whitelist += [
    "CVS_RSH",
    "ECHANGELOG_USER",
    "GPG_AGENT_INFO",
    "SSH_AGENT_PID",
    "SSH_AUTH_SOCK",
    "STY",
    "WINDOW",
    "XAUTHORITY",
]

environ_whitelist = frozenset(environ_whitelist)

environ_whitelist_re = re.compile(r"^(CCACHE_|DISTCC_).*")

# Filter selected variables in the config.environ() method so that
# they don't needlessly propagate down into the ebuild environment.
environ_filter = []

# Exclude anything that could be extremely long here (like SRC_URI)
# since that could cause execve() calls to fail with E2BIG errors. For
# example, see bug #262647.
environ_filter += [
    "DEPEND",
    "RDEPEND",
    "PDEPEND",
    "SRC_URI",
    "BDEPEND",
    "IDEPEND",
]

# misc variables inherited from the calling environment
environ_filter += [
    "INFOPATH",
    "MANPATH",
    "USER",
]

# variables that break bash
environ_filter += [
    "HISTFILE",
    "POSIXLY_CORRECT",
]

# portage config variables and variables set directly by portage
environ_filter += [
    "ACCEPT_CHOSTS",
    "ACCEPT_KEYWORDS",
    "ACCEPT_PROPERTIES",
    "ACCEPT_RESTRICT",
    "AUTOCLEAN",
    "BINPKG_COMPRESS",
    "BINPKG_COMPRESS_FLAGS",
    "CLEAN_DELAY",
    "COLLISION_IGNORE",
    "CONFIG_PROTECT",
    "CONFIG_PROTECT_MASK",
    "EGENCACHE_DEFAULT_OPTS",
    "EMERGE_DEFAULT_OPTS",
    "EMERGE_LOG_DIR",
    "EMERGE_WARNING_DELAY",
    "FETCHCOMMAND",
    "FETCHCOMMAND_FTP",
    "FETCHCOMMAND_HTTP",
    "FETCHCOMMAND_HTTPS",
    "FETCHCOMMAND_RSYNC",
    "FETCHCOMMAND_SFTP",
    "GENTOO_MIRRORS",
    "NOCONFMEM",
    "O",
    "PORTAGE_BACKGROUND",
    "PORTAGE_BACKGROUND_UNMERGE",
    "PORTAGE_BINHOST",
    "PORTAGE_BINPKG_FORMAT",
    "PORTAGE_BUILDDIR_LOCKED",
    "PORTAGE_CHECKSUM_FILTER",
    "PORTAGE_ELOG_CLASSES",
    "PORTAGE_ELOG_MAILFROM",
    "PORTAGE_ELOG_MAILSUBJECT",
    "PORTAGE_ELOG_MAILURI",
    "PORTAGE_ELOG_SYSTEM",
    "PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS",
    "PORTAGE_FETCH_RESUME_MIN_SIZE",
    "PORTAGE_GPG_DIR",
    "PORTAGE_GPG_KEY",
    "PORTAGE_GPG_SIGNING_COMMAND",
    "PORTAGE_IONICE_COMMAND",
    "PORTAGE_PACKAGE_EMPTY_ABORT",
    "PORTAGE_REPO_DUPLICATE_WARN",
    "PORTAGE_RO_DISTDIRS",
    "PORTAGE_RSYNC_EXTRA_OPTS",
    "PORTAGE_RSYNC_OPTS",
    "PORTAGE_RSYNC_RETRIES",
    "PORTAGE_SSH_OPTS",
    "PORTAGE_SYNC_STALE",
    "PORTAGE_USE",
    "PORTAGE_LOG_FILTER_FILE_CMD",
    "PORTAGE_LOGDIR",
    "PORTAGE_LOGDIR_CLEAN",
    "QUICKPKG_DEFAULT_OPTS",
    "REPOMAN_DEFAULT_OPTS",
    "RESUMECOMMAND",
    "RESUMECOMMAND_FTP",
    "RESUMECOMMAND_HTTP",
    "RESUMECOMMAND_HTTPS",
    "RESUMECOMMAND_RSYNC",
    "RESUMECOMMAND_SFTP",
    "UNINSTALL_IGNORE",
    "USE_EXPAND_HIDDEN",
    "USE_ORDER",
    "__PORTAGE_HELPER",
]

# No longer supported variables
environ_filter += ["SYNC"]

environ_filter = frozenset(environ_filter)

# Variables that are not allowed to have per-repo or per-package
# settings.
global_only_vars = frozenset(
    [
        "CONFIG_PROTECT",
    ]
)

default_globals = {
    "ACCEPT_PROPERTIES": "*",
    "PORTAGE_BZIP2_COMMAND": "bzip2",
}

validate_commands = (
    "PORTAGE_BZIP2_COMMAND",
    "PORTAGE_BUNZIP2_COMMAND",
    "PORTAGE_LOG_FILTER_FILE_CMD",
)

# To enhance usability, make some vars case insensitive
# by forcing them to lower case.
case_insensitive_vars = (
    "AUTOCLEAN",
    "NOCOLOR",
)



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class MultirepoTestCase(TestCase):
    def testMultirepo(self):
        ebuilds = {
            # Simple repo selection
            "dev-libs/A-1": {},
            "dev-libs/A-1::repo1": {},
            "dev-libs/A-2::repo1": {},
            "dev-libs/A-1::repo2": {},
            # Packages in exactly one repo
            "dev-libs/B-1": {},
            "dev-libs/C-1::repo1": {},
            # Package in repository 1 and 2, but 1 must be used
            "dev-libs/D-1::repo1": {},
            "dev-libs/D-1::repo2": {},
            "dev-libs/E-1": {},
            "dev-libs/E-1::repo1": {},
            "dev-libs/E-1::repo2": {"SLOT": "1"},
            "dev-libs/F-1::repo1": {"SLOT": "1"},
            "dev-libs/F-1::repo2": {"SLOT": "1"},
            "dev-libs/G-1::repo1": {"EAPI": "4", "IUSE": "+x +y", "REQUIRED_USE": ""},
            "dev-libs/G-1::repo2": {
                "EAPI": "4",
                "IUSE": "+x +y",
                "REQUIRED_USE": "^^ ( x y )",
            },
            "dev-libs/H-1": {
                "KEYWORDS": "x86",
                "EAPI": "3",
                "RDEPEND": "|| ( dev-libs/I:2 dev-libs/I:1 )",
            },
            "dev-libs/I-1::repo2": {"SLOT": "1"},
            "dev-libs/I-2::repo2": {"SLOT": "2"},
            "dev-libs/K-1::repo2": {},
        }

        installed = {
            "dev-libs/H-1": {
                "RDEPEND": "|| ( dev-libs/I:2 dev-libs/I:1 )",
                "EAPI": "3",
            },
            "dev-libs/I-2::repo1": {"SLOT": "2"},
            "dev-libs/K-1::repo1": {},
        }

        binpkgs = {
            "dev-libs/C-1::repo2": {},
            "dev-libs/I-2::repo1": {"SLOT": "2"},
            "dev-libs/K-1::repo2": {},
        }

        sets = {"multirepotest": ("dev-libs/A::test_repo",)}

        test_cases = (
            # Simple repo selection
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-2::repo1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A::test_repo"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A::repo2"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-1::repo2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1::repo1"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-1::repo1"],
            ),
            ResolverPlaygroundTestCase(
                ["@multirepotest"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-1"],
            ),
            # Packages in exactly one repo
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/C-1::repo1"],
            ),
            # Package in repository 1 and 2, but 2 must be used
            ResolverPlaygroundTestCase(
                ["dev-libs/D"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/D-1::repo2"],
            ),
            # --usepkg: don't reinstall on new repo without --newrepo
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                options={"--usepkg": True, "--selective": True},
                success=True,
                check_repo_names=True,
                mergelist=["[binary]dev-libs/C-1::repo2"],
            ),
            # --usepkgonly: don't reinstall on new repo without --newrepo
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                options={"--usepkgonly": True, "--selective": True},
                success=True,
                check_repo_names=True,
                mergelist=["[binary]dev-libs/C-1::repo2"],
            ),
            # --newrepo: pick ebuild if binpkg/ebuild have different repo
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                options={"--usepkg": True, "--newrepo": True, "--selective": True},
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/C-1::repo1"],
            ),
            # --newrepo --usepkgonly: ebuild is ignored
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                options={"--usepkgonly": True, "--newrepo": True, "--selective": True},
                success=True,
                check_repo_names=True,
                mergelist=["[binary]dev-libs/C-1::repo2"],
            ),
            # --newrepo: pick ebuild if binpkg/ebuild have different repo
            ResolverPlaygroundTestCase(
                ["dev-libs/I"],
                options={"--usepkg": True, "--newrepo": True, "--selective": True},
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/I-2::repo2"],
            ),
            # --newrepo --usepkgonly: if binpkg matches installed, do nothing
            ResolverPlaygroundTestCase(
                ["dev-libs/I"],
                options={"--usepkgonly": True, "--newrepo": True, "--selective": True},
                success=True,
                mergelist=[],
            ),
            # --newrepo --usepkgonly: reinstall if binpkg has new repo.
            ResolverPlaygroundTestCase(
                ["dev-libs/K"],
                options={"--usepkgonly": True, "--newrepo": True, "--selective": True},
                success=True,
                check_repo_names=True,
                mergelist=["[binary]dev-libs/K-1::repo2"],
            ),
            # --usepkgonly: don't reinstall on new repo without --newrepo.
            ResolverPlaygroundTestCase(
                ["dev-libs/K"],
                options={"--usepkgonly": True, "--selective": True},
                success=True,
                mergelist=[],
            ),
            # Atoms with slots
            ResolverPlaygroundTestCase(
                ["dev-libs/E"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/E-1::repo2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/E:1::repo2"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/E-1::repo2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/E:1"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/E-1::repo2"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/F:1"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/F-1::repo2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/F-1:1"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/F-1::repo2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/F-1:1::repo1"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/F-1::repo1"],
            ),
            # Dependency on installed dev-libs/C-2 ebuild for which ebuild is
            # not available from the same repo should not unnecessarily
            # reinstall the same version from a different repo.
            ResolverPlaygroundTestCase(
                ["dev-libs/H"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
            # Dependency on installed dev-libs/I-2 ebuild should trigger reinstall
            # when --newrepo flag is used.
            ResolverPlaygroundTestCase(
                ["dev-libs/H"],
                options={"--update": True, "--deep": True, "--newrepo": True},
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/I-2::repo2"],
            ),
            # Check interaction between repo priority and unsatisfied
            # REQUIRED_USE, for bug #350254.
            ResolverPlaygroundTestCase(
                ["=dev-libs/G-1"], check_repo_names=True, success=False
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    sets=sets,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()

    def testMultirepoUserConfig(self):
        ebuilds = {
            # package.use test
            "dev-libs/A-1": {"IUSE": "foo"},
            "dev-libs/A-2::repo1": {"IUSE": "foo"},
            "dev-libs/A-3::repo2": {},
            "dev-libs/B-1": {"DEPEND": "dev-libs/A", "EAPI": 2},
            "dev-libs/B-2": {"DEPEND": "dev-libs/A[foo]", "EAPI": 2},
            "dev-libs/B-3": {"DEPEND": "dev-libs/A[-foo]", "EAPI": 2},
            # package.accept_keywords test
            "dev-libs/C-1": {"KEYWORDS": "~x86"},
            "dev-libs/C-1::repo1": {"KEYWORDS": "~x86"},
            # package.license
            "dev-libs/D-1": {"LICENSE": "TEST"},
            "dev-libs/D-1::repo1": {"LICENSE": "TEST"},
            # package.mask
            "dev-libs/E-1": {},
            "dev-libs/E-1::repo1": {},
            "dev-libs/H-1": {},
            "dev-libs/H-1::repo1": {},
            "dev-libs/I-1::repo2": {"SLOT": "1"},
            "dev-libs/I-2::repo2": {"SLOT": "2"},
            "dev-libs/J-1": {
                "KEYWORDS": "x86",
                "EAPI": "3",
                "RDEPEND": "|| ( dev-libs/I:2 dev-libs/I:1 )",
            },
            # package.properties
            "dev-libs/F-1": {"PROPERTIES": "bar"},
            "dev-libs/F-1::repo1": {"PROPERTIES": "bar"},
            # package.unmask
            "dev-libs/G-1": {},
            "dev-libs/G-1::repo1": {},
            # package.mask with wildcards
            "dev-libs/Z-1::repo3": {},
        }

        installed = {
            "dev-libs/J-1": {
                "RDEPEND": "|| ( dev-libs/I:2 dev-libs/I:1 )",
                "EAPI": "3",
            },
            "dev-libs/I-2::repo1": {"SLOT": "2"},
        }

        user_config = {
            "package.use": ("dev-libs/A::repo1 foo",),
            "package.accept_keywords": ("=dev-libs/C-1::test_repo",),
            "package.license": ("=dev-libs/D-1::test_repo TEST",),
            "package.mask": (
                "dev-libs/E::repo1",
                "dev-libs/H",
                "dev-libs/I::repo1",
                # needed for package.unmask test
                "dev-libs/G",
                # wildcard test
                "*/*::repo3",
            ),
            "package.properties": ("dev-libs/F::repo1 -bar",),
            "package.unmask": ("dev-libs/G::test_repo",),
        }

        test_cases = (
            # package.use test
            ResolverPlaygroundTestCase(
                ["=dev-libs/B-1"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-3::repo2", "dev-libs/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/B-2"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/A-2::repo1", "dev-libs/B-2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/B-3"],
                options={"--autounmask": "n"},
                success=False,
                check_repo_names=True,
            ),
            # package.accept_keywords test
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/C-1"],
            ),
            # package.license test
            ResolverPlaygroundTestCase(
                ["dev-libs/D"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/D-1"],
            ),
            # package.mask test
            ResolverPlaygroundTestCase(
                ["dev-libs/E"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/E-1"],
            ),
            # Dependency on installed dev-libs/C-2 ebuild for which ebuild is
            # masked from the same repo should not unnecessarily pull
            # in a different slot. It should just pull in the same slot from
            # a different repo (bug #351828).
            ResolverPlaygroundTestCase(
                ["dev-libs/J"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["dev-libs/I-2"],
            ),
            # package.properties test
            ResolverPlaygroundTestCase(
                ["dev-libs/F"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/F-1"],
            ),
            # package.mask test
            ResolverPlaygroundTestCase(
                ["dev-libs/G"],
                success=True,
                check_repo_names=True,
                mergelist=["dev-libs/G-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/H"], options={"--autounmask": "n"}, success=False
            ),
            # package.mask with wildcards
            ResolverPlaygroundTestCase(
                ["dev-libs/Z"], options={"--autounmask": "n"}, success=False
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                user_config["make.conf"] = ('BINPKG_FORMAT="%s"' % binpkg_format,)
                playground = ResolverPlayground(
                    ebuilds=ebuilds, installed=installed, user_config=user_config
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 2013-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import copy
import io
import logging
import shelve
import time

from portage import os
from portage.package.ebuild.fetch import MirrorLayoutConfig
from portage.util import grabdict, grablines
from .ContentDB import ContentDB

logger = logging.getLogger(__name__)


class Config:
    def __init__(self, options, portdb, event_loop):
        self.options = options
        self.portdb = portdb
        self.event_loop = event_loop
        self.added_byte_count = 0
        self.added_file_count = 0
        self.scheduled_deletion_count = 0
        self.delete_count = 0
        self.file_owners = {}
        self.file_failures = {}
        self.start_time = time.time()
        self._open_files = []

        self.log_success = self._open_log(
            "success", getattr(options, "success_log", None), "a"
        )
        self.log_failure = self._open_log(
            "failure", getattr(options, "failure_log", None), "a"
        )

        self.distfiles = None
        if getattr(options, "distfiles", None) is not None:
            self.distfiles = options.distfiles

        self.mirrors = copy.copy(portdb.settings.thirdpartymirrors())

        if getattr(options, "mirror_overrides", None) is not None:
            self.mirrors.update(grabdict(options.mirror_overrides))

        if getattr(options, "mirror_skip", None) is not None:
            for x in options.mirror_skip.split(","):
                self.mirrors[x] = []

        self.whitelist = None
        if getattr(options, "whitelist_from", None) is not None:
            self.whitelist = set()
            for filename in options.whitelist_from:
                for line in grablines(filename):
                    line = line.strip()
                    if line and not line.startswith("#"):
                        self.whitelist.add(line)

        self.restrict_mirror_exemptions = None
        if getattr(options, "restrict_mirror_exemptions", None) is not None:
            self.restrict_mirror_exemptions = frozenset(
                options.restrict_mirror_exemptions.split(",")
            )

        self.recycle_db = None
        if getattr(options, "recycle_db", None) is not None:
            self.recycle_db = self._open_shelve(options.recycle_db, "recycle")

        self.distfiles_db = None
        if getattr(options, "distfiles_db", None) is not None:
            self.distfiles_db = self._open_shelve(options.distfiles_db, "distfiles")

        self.content_db = None
        if getattr(options, "content_db", None) is not None:
            self.content_db = ContentDB(
                self._open_shelve(options.content_db, "content")
            )

        self.deletion_db = None
        if getattr(options, "deletion_db", None) is not None:
            self.deletion_db = self._open_shelve(options.deletion_db, "deletion")

        self.layout_conf = MirrorLayoutConfig()
        if getattr(options, "layout_conf", None) is None:
            options.layout_conf = os.path.join(self.distfiles, "layout.conf")
        self.layout_conf.read_from_file(options.layout_conf)
        self.layouts = self.layout_conf.get_all_layouts()

    def _open_log(self, log_desc, log_path, mode):

        if log_path is None or getattr(self.options, "dry_run", False):
            log_func = logger.info
            line_format = "%s: %%s" % log_desc
            add_newline = False
            if log_path is not None:
                logger.warning("dry-run: %s log redirected to logging.info" % log_desc)
        else:
            self._open_files.append(io.open(log_path, mode=mode, encoding="utf_8"))
            line_format = "%s\n"
            log_func = self._open_files[-1].write

        return self._LogFormatter(line_format, log_func)

    class _LogFormatter:

        __slots__ = ("_line_format", "_log_func")

        def __init__(self, line_format, log_func):
            self._line_format = line_format
            self._log_func = log_func

        def __call__(self, msg):
            self._log_func(self._line_format % (msg,))

    def _open_shelve(self, db_file, db_desc):
        dry_run = getattr(self.options, "dry_run", False)
        if dry_run:
            open_flag = "r"
        else:
            open_flag = "c"

        if dry_run and not os.path.exists(db_file):
            db = {}
        else:
            try:
                db = shelve.open(db_file, flag=open_flag)
            except ImportError as e:
                # ImportError has different attributes for python2 vs. python3
                if (
                    getattr(e, "name", None) == "bsddb"
                    or getattr(e, "message", None) == "No module named bsddb"
                ):
                    from bsddb3 import dbshelve

                    db = dbshelve.open(db_file, flags=open_flag)

        if dry_run:
            logger.warning("dry-run: %s db opened in readonly mode" % db_desc)
            if not isinstance(db, dict):
                volatile_db = dict((k, db[k]) for k in db)
                db.close()
                db = volatile_db
        else:
            self._open_files.append(db)

        return db

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        while self._open_files:
            self._open_files.pop().close()



# elog/mod_syslog.py - elog dispatch module
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import syslog
from portage.const import EBUILD_PHASES

_pri = {
    "INFO": syslog.LOG_INFO,
    "WARN": syslog.LOG_WARNING,
    "ERROR": syslog.LOG_ERR,
    "LOG": syslog.LOG_NOTICE,
    "QA": syslog.LOG_WARNING,
}


def process(mysettings, key, logentries, fulltext):
    syslog.openlog(
        "portage",
        syslog.LOG_ERR | syslog.LOG_WARNING | syslog.LOG_INFO | syslog.LOG_NOTICE,
        syslog.LOG_LOCAL5,
    )
    for phase in EBUILD_PHASES:
        if not phase in logentries:
            continue
        for msgtype, msgcontent in logentries[phase]:
            if isinstance(msgcontent, str):
                msgcontent = [msgcontent]
            for line in msgcontent:
                line = "%s: %s: %s" % (key, phase, line)
                syslog.syslog(_pri[msgtype], line.rstrip("\n"))
    syslog.closelog()



# Copyright 2010-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ["digestgen"]

import errno

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.package.ebuild._spawn_nofetch:spawn_nofetch",
)

from portage import os
from portage.const import MANIFEST2_HASH_DEFAULTS
from portage.dbapi.porttree import FetchlistDict
from portage.dep import use_reduce
from portage.exception import (
    InvalidDependString,
    FileNotFound,
    PermissionDenied,
    PortagePackageException,
)
from portage.localization import _
from portage.output import colorize
from portage.package.ebuild.fetch import fetch
from portage.util import writemsg, writemsg_stdout
from portage.versions import catsplit


def digestgen(myarchives=None, mysettings=None, myportdb=None):
    """
    Generates a digest file if missing. Fetches files if necessary.
    NOTE: myarchives and mysettings used to be positional arguments,
            so their order must be preserved for backward compatibility.
    @param mysettings: the ebuild config (mysettings["O"] must correspond
            to the ebuild's parent directory)
    @type mysettings: config
    @param myportdb: a portdbapi instance
    @type myportdb: portdbapi
    @rtype: int
    @return: 1 on success and 0 on failure
    """
    if mysettings is None or myportdb is None:
        raise TypeError(
            "portage.digestgen(): 'mysettings' and 'myportdb' parameter are required."
        )

    try:
        portage._doebuild_manifest_exempt_depend += 1
        distfiles_map = {}
        fetchlist_dict = FetchlistDict(mysettings["O"], mysettings, myportdb)
        for cpv in fetchlist_dict:
            try:
                for myfile in fetchlist_dict[cpv]:
                    distfiles_map.setdefault(myfile, []).append(cpv)
            except InvalidDependString as e:
                writemsg("!!! %s\n" % str(e), noiselevel=-1)
                del e
                return 0
        mytree = os.path.dirname(os.path.dirname(mysettings["O"]))
        try:
            mf = mysettings.repositories.get_repo_for_location(mytree)
        except KeyError:
            # backward compatibility
            mytree = os.path.realpath(mytree)
            mf = mysettings.repositories.get_repo_for_location(mytree)

        repo_required_hashes = mf.manifest_required_hashes
        if repo_required_hashes is None:
            repo_required_hashes = MANIFEST2_HASH_DEFAULTS
        mf = mf.load_manifest(
            mysettings["O"], mysettings["DISTDIR"], fetchlist_dict=fetchlist_dict
        )

        if not mf.allow_create:
            writemsg_stdout(
                _(
                    ">>> Skipping creating Manifest for %s; "
                    "repository is configured to not use them\n"
                )
                % mysettings["O"]
            )
            return 1

        # Don't require all hashes since that can trigger excessive
        # fetches when sufficient digests already exist.  To ease transition
        # while Manifest 1 is being removed, only require hashes that will
        # exist before and after the transition.
        required_hash_types = set()
        required_hash_types.add("size")
        required_hash_types.update(repo_required_hashes)
        dist_hashes = mf.fhashdict.get("DIST", {})

        # To avoid accidental regeneration of digests with the incorrect
        # files (such as partially downloaded files), trigger the fetch
        # code if the file exists and it's size doesn't match the current
        # manifest entry. If there really is a legitimate reason for the
        # digest to change, `ebuild --force digest` can be used to avoid
        # triggering this code (or else the old digests can be manually
        # removed from the Manifest).
        missing_files = []
        for myfile in distfiles_map:
            myhashes = dist_hashes.get(myfile)
            if not myhashes:
                try:
                    st = os.stat(os.path.join(mysettings["DISTDIR"], myfile))
                except OSError:
                    st = None
                if st is None or st.st_size == 0:
                    missing_files.append(myfile)
                continue
            size = myhashes.get("size")

            try:
                st = os.stat(os.path.join(mysettings["DISTDIR"], myfile))
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                del e
                if size == 0:
                    missing_files.append(myfile)
                    continue
                if required_hash_types.difference(myhashes):
                    missing_files.append(myfile)
                    continue
            else:
                if st.st_size == 0 or size is not None and size != st.st_size:
                    missing_files.append(myfile)
                    continue

        for myfile in missing_files:
            uris = set()
            all_restrict = set()
            for cpv in distfiles_map[myfile]:
                uris.update(myportdb.getFetchMap(cpv, mytree=mytree)[myfile])
                restrict = myportdb.aux_get(cpv, ["RESTRICT"], mytree=mytree)[0]
                # Here we ignore conditional parts of RESTRICT since
                # they don't apply unconditionally. Assume such
                # conditionals only apply on the client side where
                # digestgen() does not need to be called.
                all_restrict.update(use_reduce(restrict, flat=True, matchnone=True))

                # fetch() uses CATEGORY and PF to display a message
                # when fetch restriction is triggered.
                cat, pf = catsplit(cpv)
                mysettings["CATEGORY"] = cat
                mysettings["PF"] = pf

            # fetch() uses PORTAGE_RESTRICT to control fetch
            # restriction, which is only applied to files that
            # are not fetchable via a mirror:// URI.
            mysettings["PORTAGE_RESTRICT"] = " ".join(all_restrict)

            try:
                st = os.stat(os.path.join(mysettings["DISTDIR"], myfile))
            except OSError:
                st = None

            if not fetch({myfile: uris}, mysettings):
                myebuild = os.path.join(mysettings["O"], catsplit(cpv)[1] + ".ebuild")
                spawn_nofetch(myportdb, myebuild)
                writemsg(
                    _("!!! Fetch failed for %s, can't update Manifest\n") % myfile,
                    noiselevel=-1,
                )
                if myfile in dist_hashes and st is not None and st.st_size > 0:
                    # stat result is obtained before calling fetch(),
                    # since fetch may rename the existing file if the
                    # digest does not match.
                    cmd = colorize(
                        "INFORM",
                        "ebuild --force %s manifest" % os.path.basename(myebuild),
                    )
                    writemsg(
                        (
                            _(
                                "!!! If you would like to forcefully replace the existing Manifest entry\n"
                                "!!! for %s, use the following command:\n"
                            )
                            % myfile
                        )
                        + "!!!    %s\n" % cmd,
                        noiselevel=-1,
                    )
                return 0

        writemsg_stdout(_(">>> Creating Manifest for %s\n") % mysettings["O"])
        try:
            mf.create(
                assumeDistHashesSometimes=True,
                assumeDistHashesAlways=("assume-digests" in mysettings.features),
            )
        except FileNotFound as e:
            writemsg(
                _("!!! File %s doesn't exist, can't update Manifest\n") % e,
                noiselevel=-1,
            )
            return 0
        except PortagePackageException as e:
            writemsg(("!!! %s\n") % (e,), noiselevel=-1)
            return 0
        try:
            mf.write(sign=False)
        except PermissionDenied as e:
            writemsg(_("!!! Permission Denied: %s\n") % (e,), noiselevel=-1)
            return 0
        if "assume-digests" not in mysettings.features:
            distlist = list(mf.fhashdict.get("DIST", {}))
            distlist.sort()
            auto_assumed = []
            for filename in distlist:
                if not os.path.exists(os.path.join(mysettings["DISTDIR"], filename)):
                    auto_assumed.append(filename)
            if auto_assumed:
                cp = os.path.sep.join(mysettings["O"].split(os.path.sep)[-2:])
                pkgs = myportdb.cp_list(cp, mytree=mytree)
                pkgs.sort()
                writemsg_stdout(
                    "  digest.assumed"
                    + colorize("WARN", str(len(auto_assumed)).rjust(18))
                    + "\n"
                )
                for pkg_key in pkgs:
                    fetchlist = myportdb.getFetchMap(pkg_key, mytree=mytree)
                    pv = pkg_key.split("/")[1]
                    for filename in auto_assumed:
                        if filename in fetchlist:
                            writemsg_stdout("   %s::%s\n" % (pv, filename))
        return 1
    finally:
        portage._doebuild_manifest_exempt_depend -= 1



# Copright Gentoo Foundation 2006-2020
# Portage Unit Testing Functionality

import io
import tarfile
import tempfile
from os import urandom

from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.gpkg import gpkg
from portage.exception import (
    InvalidBinaryPackageFormat,
    DigestException,
    MissingSignature,
)


class test_gpkg_checksum_case(TestCase):
    def test_gpkg_missing_header(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name != os.path.join("test", binpkg_1.gpkg_version):
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                InvalidBinaryPackageFormat,
                binpkg_2.decompress,
                os.path.join(tmpdir, "test"),
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_missing_manifest(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name != os.path.join("test", "Manifest"):
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                MissingSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_missing_files(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data2"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if "image.tar" not in f.name:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                DigestException, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_extra_files(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        tar_2.addfile(f, tar_1.extractfile(f))
                    data_tarinfo = tarfile.TarInfo(os.path.join("test", "data2"))
                    data_tarinfo.size = len(data)
                    data2 = io.BytesIO(data)
                    tar_2.addfile(data_tarinfo, data2)
                    data2.close()

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                DigestException, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_incorrect_checksum(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == os.path.join("test", "Manifest"):
                            data = io.BytesIO(tar_1.extractfile(f).read())
                            data_view = data.getbuffer()
                            data_view[-16:] = b"20a6d80ab0320fh9"
                            del data_view
                            tar_2.addfile(f, data)
                            data.close()
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                DigestException, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_duplicate_files(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(100)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        tar_2.addfile(f, tar_1.extractfile(f))
                        tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                InvalidBinaryPackageFormat,
                binpkg_2.decompress,
                os.path.join(tmpdir, "test"),
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_manifest_duplicate_files(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(100)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == os.path.join("test", "Manifest"):
                            manifest = tar_1.extractfile(f).read()
                            data = io.BytesIO(manifest)
                            data.seek(io.SEEK_END)
                            data.write(b"\n")
                            data.write(manifest)
                            f.size = data.tell()
                            data.seek(0)
                            tar_2.addfile(f, data)
                            data.close()
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                DigestException, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_different_size_file(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature -gpg-keepalive"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(100)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if "image" in f.name:
                            data = tar_1.extractfile(f).read()
                            data = data + b"1234"
                            f.size = len(data)
                            tar_2.addfile(f, io.BytesIO(data))
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                DigestException,
                binpkg_2.decompress,
                os.path.join(tmpdir, "test"),
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import cpvequal
from portage.exception import PortageException


class TestStandalone(TestCase):
    """Test some small functions portage.dep"""

    def testCPVequal(self):

        test_cases = (
            ("sys-apps/portage-2.1", "sys-apps/portage-2.1", True),
            ("sys-apps/portage-2.1", "sys-apps/portage-2.0", False),
            ("sys-apps/portage-2.1", "sys-apps/portage-2.1-r1", False),
            ("sys-apps/portage-2.1-r1", "sys-apps/portage-2.1", False),
            ("sys-apps/portage-2.1_alpha3", "sys-apps/portage-2.1", False),
            ("sys-apps/portage-2.1_alpha3_p6", "sys-apps/portage-2.1_alpha3", False),
            ("sys-apps/portage-2.1_alpha3", "sys-apps/portage-2.1", False),
            ("sys-apps/portage-2.1", "sys-apps/X-2.1", False),
            ("sys-apps/portage-2.1", "portage-2.1", False),
        )

        test_cases_xfail = (
            ("sys-apps/portage", "sys-apps/portage"),
            ("sys-apps/portage-2.1-6", "sys-apps/portage-2.1-6"),
        )

        for cpv1, cpv2, expected_result in test_cases:
            self.assertEqual(
                cpvequal(cpv1, cpv2),
                expected_result,
                "cpvequal('%s', '%s') != %s" % (cpv1, cpv2, expected_result),
            )

        for cpv1, cpv2 in test_cases_xfail:
            self.assertRaisesMsg(
                "cpvequal(%s, %s)" % (cpv1, cpv2),
                PortageException,
                cpvequal,
                cpv1,
                cpv2,
            )



# Copyright 2013-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections
import errno
import logging
import random
import subprocess

import portage
from portage import _encodings, _unicode_encode
from portage import os
from portage.util import ensure_dirs
from portage.util._async.FileCopier import FileCopier
from portage.util._async.FileDigester import FileDigester
from portage.util._async.PipeLogger import PipeLogger
from portage.util._async.PopenProcess import PopenProcess
from _emerge.CompositeTask import CompositeTask

logger = logging.getLogger(__name__)

default_hash_name = portage.const.MANIFEST2_HASH_DEFAULT

# Use --no-check-certificate since Manifest digests should provide
# enough security, and certificates can be self-signed or whatnot.
default_fetchcommand = 'wget -c -v -t 1 --passive-ftp --no-check-certificate --timeout=60 -O "${DISTDIR}/${FILE}" "${URI}"'


class FetchTask(CompositeTask):

    __slots__ = (
        "distfile",
        "digests",
        "config",
        "cpv",
        "restrict",
        "uri_tuple",
        "_current_mirror",
        "_current_stat",
        "_fetch_tmp_dir_info",
        "_fetch_tmp_file",
        "_fs_mirror_stack",
        "_mirror_stack",
        "_previously_added",
        "_primaryuri_stack",
        "_log_path",
        "_tried_uris",
    )

    def _start(self):

        if (
            self.config.options.fetch_log_dir is not None
            and not self.config.options.dry_run
        ):
            self._log_path = os.path.join(
                self.config.options.fetch_log_dir, self.distfile + ".log"
            )

        self._previously_added = True
        if (
            self.config.distfiles_db is not None
            and self.distfile not in self.config.distfiles_db
        ):
            self._previously_added = False
            # Convert _pkg_str to str in order to prevent pickle problems.
            self.config.distfiles_db[self.distfile] = str(self.cpv)

        if self.config.content_db is not None:
            self.config.content_db.add(self.distfile)

        if not self._have_needed_digests():
            msg = "incomplete digests: %s" % " ".join(self.digests)
            self.scheduler.output(
                msg, background=self.background, log_path=self._log_path
            )
            self.config.log_failure("%s\t%s\t%s" % (self.cpv, self.distfile, msg))
            self.config.file_failures[self.distfile] = self.cpv
            self.returncode = os.EX_OK
            self._async_wait()
            return

        st = None
        for layout in self.config.layouts:
            distfile_path = os.path.join(
                self.config.options.distfiles, layout.get_path(self.distfile)
            )
            try:
                st = os.stat(distfile_path)
            except OSError as e:
                if e.errno not in (errno.ENOENT, errno.ESTALE):
                    msg = "%s stat failed in %s: %s" % (self.distfile, "distfiles", e)
                    self.scheduler.output(
                        msg + "\n", background=True, log_path=self._log_path
                    )
                    logger.error(msg)
            else:
                break

        size_ok = st is not None and st.st_size == self.digests["size"]

        if not size_ok:
            if self.config.options.dry_run:
                if st is not None:
                    logger.info(
                        ("dry-run: delete '%s' with " "wrong size from distfiles")
                        % (self.distfile,)
                    )
            else:
                # Do the unlink in order to ensure that the path is clear,
                # even if stat raised ENOENT, since a broken symlink can
                # trigger ENOENT.
                unlink_success = True
                for layout in self.config.layouts:
                    unlink_path = os.path.join(
                        self.config.options.distfiles, layout.get_path(self.distfile)
                    )
                    if self._unlink_file(unlink_path, "distfiles"):
                        if st is not None:
                            logger.debug(
                                ("delete '%s' with " "wrong size from distfiles")
                                % (self.distfile,)
                            )
                    else:
                        self.config.log_failure(
                            "%s\t%s\t%s"
                            % (self.cpv, self.distfile, "unlink failed in distfiles")
                        )
                        unlink_success = False
                if not unlink_success:
                    self.returncode = os.EX_OK
                    self._async_wait()
                    return

        if size_ok:
            if self.config.options.verify_existing_digest:
                self._start_task(
                    FileDigester(
                        file_path=distfile_path,
                        hash_names=(self._select_hash(),),
                        background=self.background,
                        logfile=self._log_path,
                    ),
                    self._distfiles_digester_exit,
                )
                return

            self._success()
            self.returncode = os.EX_OK
            self._async_wait()
            return

        self._start_fetch()

    def _success(self):
        if not self._previously_added:
            size = self.digests["size"]
            self.config.added_byte_count += size
            self.config.added_file_count += 1
            self.config.log_success(
                "%s\t%s\tadded %i bytes" % (self.cpv, self.distfile, size)
            )

        if self._log_path is not None:
            if not self.config.options.dry_run:
                try:
                    os.unlink(self._log_path)
                except OSError:
                    pass

        if self.config.options.recycle_dir is not None:

            recycle_file = os.path.join(self.config.options.recycle_dir, self.distfile)

            if self.config.options.dry_run:
                if os.path.exists(recycle_file):
                    logger.info("dry-run: delete '%s' from recycle" % (self.distfile,))
            else:
                try:
                    os.unlink(recycle_file)
                except OSError:
                    pass
                else:
                    logger.debug("delete '%s' from recycle" % (self.distfile,))

    def _distfiles_digester_exit(self, digester):

        self._assert_current(digester)
        if self._was_cancelled():
            self.wait()
            return

        if self._default_exit(digester) != os.EX_OK:
            # IOError reading file in our main distfiles directory? This
            # is a bad situation which normally does not occur, so
            # skip this file and report it, in order to draw attention
            # from the administrator.
            msg = "%s distfiles digester failed unexpectedly" % (self.distfile,)
            self.scheduler.output(msg + "\n", background=True, log_path=self._log_path)
            logger.error(msg)
            self.config.log_failure("%s\t%s\t%s" % (self.cpv, self.distfile, msg))
            self.config.file_failures[self.distfile] = self.cpv
            self.wait()
            return

        wrong_digest = self._find_bad_digest(digester.digests)
        if wrong_digest is None:
            self._success()
            self.returncode = os.EX_OK
            self.wait()
            return

        self._start_fetch()

    _mirror_info = collections.namedtuple("_mirror_info", "name location")

    def _start_fetch(self):

        self._previously_added = False
        self._fs_mirror_stack = []
        if self.config.options.distfiles_local is not None:
            self._fs_mirror_stack.append(
                self._mirror_info(
                    "distfiles-local", self.config.options.distfiles_local
                )
            )
        if self.config.options.recycle_dir is not None:
            self._fs_mirror_stack.append(
                self._mirror_info("recycle", self.config.options.recycle_dir)
            )

        self._primaryuri_stack = []
        self._mirror_stack = []
        for uri in reversed(self.uri_tuple):
            if uri.startswith("mirror://"):
                self._mirror_stack.append(
                    self._mirror_iterator(uri, self.config.mirrors)
                )
            else:
                self._primaryuri_stack.append(uri)

        self._tried_uris = set()
        self._try_next_mirror()

    @staticmethod
    def _mirror_iterator(uri, mirrors_dict):

        slash_index = uri.find("/", 9)
        if slash_index != -1:
            mirror_name = uri[9:slash_index].strip("/")
            mirrors = mirrors_dict.get(mirror_name)
            if not mirrors:
                return
            mirrors = list(mirrors)
            while mirrors:
                mirror = mirrors.pop(random.randint(0, len(mirrors) - 1))
                yield mirror.rstrip("/") + "/" + uri[slash_index + 1 :]

    def _try_next_mirror(self):
        if self._fs_mirror_stack:
            self._fetch_fs(self._fs_mirror_stack.pop())
            return
        uri = self._next_uri()
        if uri is not None:
            self._tried_uris.add(uri)
            self._fetch_uri(uri)
            return

        if self._tried_uris:
            msg = "all uris failed"
        else:
            msg = "no fetchable uris"

        self.config.log_failure("%s\t%s\t%s" % (self.cpv, self.distfile, msg))
        self.config.file_failures[self.distfile] = self.cpv
        self.returncode = os.EX_OK
        self.wait()

    def _next_uri(self):
        remaining_tries = self.config.options.tries - len(self._tried_uris)
        if remaining_tries > 0:

            if remaining_tries <= self.config.options.tries // 2:
                while self._primaryuri_stack:
                    uri = self._primaryuri_stack.pop()
                    if uri not in self._tried_uris:
                        return uri

            while self._mirror_stack:
                uri = next(self._mirror_stack[-1], None)
                if uri is None:
                    self._mirror_stack.pop()
                else:
                    if uri not in self._tried_uris:
                        return uri

            while self._primaryuri_stack:
                uri = self._primaryuri_stack.pop()
                if uri not in self._tried_uris:
                    return uri

        return None

    def _fetch_fs(self, mirror_info):
        file_path = os.path.join(mirror_info.location, self.distfile)

        st = None
        size_ok = False
        try:
            st = os.stat(file_path)
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                msg = "%s stat failed in %s: %s" % (self.distfile, mirror_info.name, e)
                self.scheduler.output(
                    msg + "\n", background=True, log_path=self._log_path
                )
                logger.error(msg)
        else:
            size_ok = st.st_size == self.digests["size"]
            self._current_stat = st

        if size_ok:
            self._current_mirror = mirror_info
            self._start_task(
                FileDigester(
                    file_path=file_path,
                    hash_names=(self._select_hash(),),
                    background=self.background,
                    logfile=self._log_path,
                ),
                self._fs_mirror_digester_exit,
            )
        else:
            self._try_next_mirror()

    def _fs_mirror_digester_exit(self, digester):

        self._assert_current(digester)
        if self._was_cancelled():
            self.wait()
            return

        current_mirror = self._current_mirror
        if digester.returncode != os.EX_OK:
            msg = "%s %s digester failed unexpectedly" % (
                self.distfile,
                current_mirror.name,
            )
            self.scheduler.output(msg + "\n", background=True, log_path=self._log_path)
            logger.error(msg)
        else:
            bad_digest = self._find_bad_digest(digester.digests)
            if bad_digest is not None:
                msg = "%s %s has bad %s digest: expected %s, got %s" % (
                    self.distfile,
                    current_mirror.name,
                    bad_digest,
                    self.digests[bad_digest],
                    digester.digests[bad_digest],
                )
                self.scheduler.output(
                    msg + "\n", background=True, log_path=self._log_path
                )
                logger.error(msg)
            elif self.config.options.dry_run:
                # Report success without actually touching any files
                if self._same_device(
                    current_mirror.location, self.config.options.distfiles
                ):
                    logger.info(
                        ("dry-run: hardlink '%s' from %s " "to distfiles")
                        % (self.distfile, current_mirror.name)
                    )
                else:
                    logger.info(
                        "dry-run: copy '%s' from %s to distfiles"
                        % (self.distfile, current_mirror.name)
                    )
                self._success()
                self.returncode = os.EX_OK
                self.wait()
                return
            else:
                src = os.path.join(current_mirror.location, self.distfile)
                dest = os.path.join(
                    self.config.options.distfiles,
                    self.config.layouts[0].get_path(self.distfile),
                )
                if self._hardlink_atomic(
                    src, dest, "%s to %s" % (current_mirror.name, "distfiles")
                ):
                    logger.debug(
                        "hardlink '%s' from %s to distfiles"
                        % (self.distfile, current_mirror.name)
                    )
                    self._success()
                    self.returncode = os.EX_OK
                    self.wait()
                    return

                self._start_task(
                    FileCopier(
                        src_path=src,
                        dest_path=dest,
                        background=(self.background and self._log_path is not None),
                        logfile=self._log_path,
                    ),
                    self._fs_mirror_copier_exit,
                )
                return

        self._try_next_mirror()

    def _fs_mirror_copier_exit(self, copier):

        self._assert_current(copier)
        if self._was_cancelled():
            self.wait()
            return

        current_mirror = self._current_mirror
        if copier.returncode != os.EX_OK:
            msg = "%s %s copy failed unexpectedly: %s" % (
                self.distfile,
                current_mirror.name,
                copier.future.exception(),
            )
            self.scheduler.output(msg + "\n", background=True, log_path=self._log_path)
            logger.error(msg)
        else:

            logger.debug(
                "copy '%s' from %s to distfiles" % (self.distfile, current_mirror.name)
            )

            # Apply the timestamp from the source file, but
            # just rely on umask for permissions.
            try:
                os.utime(
                    copier.dest_path,
                    ns=(self._current_stat.st_mtime_ns, self._current_stat.st_mtime_ns),
                )
            except OSError as e:
                msg = "%s %s utime failed unexpectedly: %s" % (
                    self.distfile,
                    current_mirror.name,
                    e,
                )
                self.scheduler.output(
                    msg + "\n", background=True, log_path=self._log_path
                )
                logger.error(msg)

            self._success()
            self.returncode = os.EX_OK
            self.wait()
            return

        self._try_next_mirror()

    def _fetch_uri(self, uri):

        if self.config.options.dry_run:
            # Simply report success.
            logger.info("dry-run: fetch '%s' from '%s'" % (self.distfile, uri))
            self._success()
            self.returncode = os.EX_OK
            self._async_wait()
            return

        if self.config.options.temp_dir:
            self._fetch_tmp_dir_info = "temp-dir"
            distdir = self.config.options.temp_dir
        else:
            self._fetch_tmp_dir_info = "distfiles"
            distdir = self.config.options.distfiles

        tmp_basename = self.distfile + "._emirrordist_fetch_.%s" % portage.getpid()

        variables = {"DISTDIR": distdir, "URI": uri, "FILE": tmp_basename}

        self._fetch_tmp_file = os.path.join(distdir, tmp_basename)

        try:
            os.unlink(self._fetch_tmp_file)
        except OSError:
            pass

        args = portage.util.shlex_split(default_fetchcommand)
        args = [portage.util.varexpand(x, mydict=variables) for x in args]

        args = [
            _unicode_encode(x, encoding=_encodings["fs"], errors="strict") for x in args
        ]

        null_fd = os.open(os.devnull, os.O_RDONLY)
        fetcher = PopenProcess(
            background=self.background,
            proc=subprocess.Popen(
                args, stdin=null_fd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            ),
            scheduler=self.scheduler,
        )
        os.close(null_fd)

        fetcher.pipe_reader = PipeLogger(
            background=self.background,
            input_fd=fetcher.proc.stdout,
            log_file_path=self._log_path,
            scheduler=self.scheduler,
        )

        self._start_task(fetcher, self._fetcher_exit)

    def _fetcher_exit(self, fetcher):

        self._assert_current(fetcher)
        if self._was_cancelled():
            self.wait()
            return

        if os.path.exists(self._fetch_tmp_file):
            self._start_task(
                FileDigester(
                    file_path=self._fetch_tmp_file,
                    hash_names=(self._select_hash(),),
                    background=self.background,
                    logfile=self._log_path,
                ),
                self._fetch_digester_exit,
            )
        else:
            self._try_next_mirror()

    def _fetch_digester_exit(self, digester):

        self._assert_current(digester)
        if self._was_cancelled():
            self.wait()
            return

        if digester.returncode != os.EX_OK:
            msg = "%s %s digester failed unexpectedly" % (
                self.distfile,
                self._fetch_tmp_dir_info,
            )
            self.scheduler.output(msg + "\n", background=True, log_path=self._log_path)
            logger.error(msg)
        else:
            bad_digest = self._find_bad_digest(digester.digests)
            if bad_digest is not None:
                msg = "%s has bad %s digest: expected %s, got %s" % (
                    self.distfile,
                    bad_digest,
                    self.digests[bad_digest],
                    digester.digests[bad_digest],
                )
                self.scheduler.output(
                    msg + "\n", background=True, log_path=self._log_path
                )
                try:
                    os.unlink(self._fetch_tmp_file)
                except OSError:
                    pass
            else:
                dest = os.path.join(
                    self.config.options.distfiles,
                    self.config.layouts[0].get_path(self.distfile),
                )
                ensure_dirs(os.path.dirname(dest))
                try:
                    os.rename(self._fetch_tmp_file, dest)
                except OSError:
                    self._start_task(
                        FileCopier(
                            src_path=self._fetch_tmp_file,
                            dest_path=dest,
                            background=(self.background and self._log_path is not None),
                            logfile=self._log_path,
                        ),
                        self._fetch_copier_exit,
                    )
                    return
                else:
                    self._make_layout_links()
                    return

        self._try_next_mirror()

    def _fetch_copier_exit(self, copier):

        self._assert_current(copier)

        try:
            os.unlink(self._fetch_tmp_file)
        except OSError:
            pass

        if self._was_cancelled():
            self.wait()
            return

        if copier.returncode == os.EX_OK:
            self._make_layout_links()
        else:
            # out of space?
            msg = "%s %s copy failed unexpectedly: %s" % (
                self.distfile,
                self._fetch_tmp_dir_info,
                copier.future.exception(),
            )
            self.scheduler.output(msg + "\n", background=True, log_path=self._log_path)
            logger.error(msg)
            self.config.log_failure("%s\t%s\t%s" % (self.cpv, self.distfile, msg))
            self.config.file_failures[self.distfile] = self.cpv
            self.returncode = 1
            self.wait()

    def _make_layout_links(self):
        dist_path = None
        success = True
        for layout in self.config.layouts[1:]:
            if dist_path is None:
                dist_path = os.path.join(
                    self.config.options.distfiles,
                    self.config.layouts[0].get_path(self.distfile),
                )
            link_path = os.path.join(
                self.config.options.distfiles, layout.get_path(self.distfile)
            )
            ensure_dirs(os.path.dirname(link_path))
            src_path = dist_path
            if self.config.options.symlinks:
                src_path = os.path.relpath(dist_path, os.path.dirname(link_path))

            if not self._hardlink_atomic(
                src_path,
                link_path,
                "%s -> %s" % (link_path, src_path),
                self.config.options.symlinks,
            ):
                success = False
                break

        if success:
            self._success()
            self.returncode = os.EX_OK
        else:
            msg = "failed to create distfiles layout {}".format(
                "symlink" if self.config.options.symlinks else "hardlink"
            )
            self.config.log_failure("%s\t%s\t%s" % (self.cpv, self.distfile, msg))
            self.config.file_failures[self.distfile] = self.cpv
            self.returncode = 1

        self.wait()

    def _unlink_file(self, file_path, dir_info):
        try:
            os.unlink(file_path)
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                msg = "unlink '%s' failed in %s: %s" % (self.distfile, dir_info, e)
                self.scheduler.output(
                    msg + "\n", background=True, log_path=self._log_path
                )
                logger.error(msg)
                return False
        return True

    def _have_needed_digests(self):
        return "size" in self.digests and self._select_hash() is not None

    def _select_hash(self):
        if default_hash_name in self.digests:
            return default_hash_name
        for hash_name in self.digests:
            if (
                hash_name != "size"
                and hash_name in portage.checksum.get_valid_checksum_keys()
            ):
                return hash_name

        return None

    def _find_bad_digest(self, digests):
        for hash_name, hash_value in digests.items():
            if self.digests[hash_name] != hash_value:
                return hash_name
        return None

    @staticmethod
    def _same_device(path1, path2):
        try:
            st1 = os.stat(path1)
            st2 = os.stat(path2)
        except OSError:
            return False
        else:
            return st1.st_dev == st2.st_dev

    def _hardlink_atomic(self, src, dest, dir_info, symlink=False):

        head, tail = os.path.split(dest)
        hardlink_tmp = os.path.join(
            head, ".%s._mirrordist_hardlink_.%s" % (tail, portage.getpid())
        )

        try:
            try:
                if symlink:
                    os.symlink(src, hardlink_tmp)
                else:
                    os.link(src, hardlink_tmp)
            except OSError as e:
                if e.errno != errno.EXDEV:
                    msg = "hardlink %s from %s failed: %s" % (
                        self.distfile,
                        dir_info,
                        e,
                    )
                    self.scheduler.output(
                        msg + "\n", background=True, log_path=self._log_path
                    )
                    logger.error(msg)
                return False

            try:
                os.rename(hardlink_tmp, dest)
            except OSError as e:
                msg = "hardlink rename '%s' from %s failed: %s" % (
                    self.distfile,
                    dir_info,
                    e,
                )
                self.scheduler.output(
                    msg + "\n", background=True, log_path=self._log_path
                )
                logger.error(msg)
                return False
        finally:
            try:
                os.unlink(hardlink_tmp)
            except OSError:
                pass

        return True



# Copyright 2014-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import io
import functools
import operator
import os

import portage
from portage import _encodings
from portage.dep import Atom
from portage.exception import FileNotFound
from portage.cache.index.IndexStreamIterator import IndexStreamIterator
from portage.cache.index.pkg_desc_index import (
    pkg_desc_index_line_read,
    pkg_desc_index_node,
)
from portage.util.iterators.MultiIterGroupBy import MultiIterGroupBy


class IndexedPortdb:
    """
    A portdbapi interface that uses a package description index to
    improve performance. If the description index is missing for a
    particular repository, then all metadata for that repository is
    obtained using the normal pordbapi.aux_get method.

    For performance reasons, the match method only supports package
    name and version constraints. For the same reason, the xmatch
    method is not implemented.
    """

    # Match returns unordered results.
    match_unordered = True

    _copy_attrs = (
        "cpv_exists",
        "findname",
        "getFetchMap",
        "_aux_cache_keys",
        "_cpv_sort_ascending",
        "_have_root_eclass_dir",
    )

    def __init__(self, portdb):

        self._portdb = portdb

        for k in self._copy_attrs:
            setattr(self, k, getattr(portdb, k))

        self._desc_cache = None
        self._cp_map = None
        self._unindexed_cp_map = None

    def _init_index(self):

        cp_map = {}
        desc_cache = {}
        self._desc_cache = desc_cache
        self._cp_map = cp_map
        index_missing = []

        streams = []
        for repo_path in self._portdb.porttrees:
            outside_repo = os.path.join(
                self._portdb.depcachedir, repo_path.lstrip(os.sep)
            )
            filenames = []
            for parent_dir in (repo_path, outside_repo):
                filenames.append(os.path.join(parent_dir, "metadata", "pkg_desc_index"))

            repo_name = self._portdb.getRepositoryName(repo_path)

            try:
                f = None
                for filename in filenames:
                    try:
                        f = io.open(filename, encoding=_encodings["repo.content"])
                    except IOError as e:
                        if e.errno not in (errno.ENOENT, errno.ESTALE):
                            raise
                    else:
                        break

                if f is None:
                    raise FileNotFound(filename)

                streams.append(
                    iter(
                        IndexStreamIterator(
                            f,
                            functools.partial(pkg_desc_index_line_read, repo=repo_name),
                        )
                    )
                )
            except FileNotFound:
                index_missing.append(repo_path)

        if index_missing:
            self._unindexed_cp_map = {}

            class _NonIndexedStream:
                def __iter__(self_):
                    for cp in self._portdb.cp_all(trees=index_missing):
                        # Don't call cp_list yet, since it's a waste
                        # if the package name does not match the current
                        # search.
                        self._unindexed_cp_map[cp] = index_missing
                        yield pkg_desc_index_node(cp, (), None)

            streams.append(iter(_NonIndexedStream()))

        if streams:
            if len(streams) == 1:
                cp_group_iter = ([node] for node in streams[0])
            else:
                cp_group_iter = MultiIterGroupBy(streams, key=operator.attrgetter("cp"))

            for cp_group in cp_group_iter:

                new_cp = None
                cp_list = cp_map.get(cp_group[0].cp)
                if cp_list is None:
                    new_cp = cp_group[0].cp
                    cp_list = []
                    cp_map[cp_group[0].cp] = cp_list

                for entry in cp_group:
                    cp_list.extend(entry.cpv_list)
                    if entry.desc is not None:
                        for cpv in entry.cpv_list:
                            desc_cache[cpv] = entry.desc

                if new_cp is not None:
                    yield cp_group[0].cp

    def cp_all(self, sort=True):
        """
        Returns an ordered iterator instead of a list, so that search
        results can be displayed incrementally.
        """
        if self._cp_map is None:
            return self._init_index()
        return iter(sorted(self._cp_map)) if sort else iter(self._cp_map)

    def match(self, atom):
        """
        For performance reasons, only package name and version
        constraints are supported, and the returned list is
        unordered.
        """
        if not isinstance(atom, Atom):
            atom = Atom(atom)
        cp_list = self._cp_map.get(atom.cp)
        if cp_list is None:
            return []

        if self._unindexed_cp_map is not None:
            try:
                unindexed = self._unindexed_cp_map.pop(atom.cp)
            except KeyError:
                pass
            else:
                cp_list.extend(self._portdb.cp_list(atom.cp, mytree=unindexed))

        if atom == atom.cp:
            return cp_list[:]
        return portage.match_from_list(atom, cp_list)

    def aux_get(self, cpv, attrs, myrepo=None):
        if len(attrs) == 1 and attrs[0] == "DESCRIPTION":
            try:
                return [self._desc_cache[cpv]]
            except KeyError:
                pass
        return self._portdb.aux_get(cpv, attrs)



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Brian Harring (ferringb@gentoo.org)

import dbm

try:
    import dbm.gnu as gdbm
except ImportError:
    gdbm = None

import pickle

from portage import _unicode_encode
from portage import os
from portage.cache import fs_template
from portage.cache import cache_errors


class database(fs_template.FsBased):

    validation_chf = "md5"
    chf_types = ("md5", "mtime")

    autocommits = True
    cleanse_keys = True
    serialize_eclasses = False

    def __init__(self, *args, **config):
        super(database, self).__init__(*args, **config)

        default_db = config.get("dbtype", "anydbm")
        if not default_db.startswith("."):
            default_db = "." + default_db

        self._db_path = os.path.join(
            self.location, fs_template.gen_label(self.location, self.label) + default_db
        )
        self.__db = None
        mode = "w"
        if dbm.whichdb(self._db_path) in ("dbm.gnu", "gdbm"):
            # Allow multiple concurrent writers (see bug #53607).
            mode += "u"
        try:
            # dbm.open() will not work with bytes in python-3.1:
            #   TypeError: can't concat bytes to str
            self.__db = dbm.open(self._db_path, mode, self._perms)
        except dbm.error:
            # XXX handle this at some point
            try:
                self._ensure_dirs()
                self._ensure_dirs(self._db_path)
            except (OSError, IOError) as e:
                raise cache_errors.InitializationError(self.__class__, e)

            # try again if failed
            try:
                if self.__db is None:
                    # dbm.open() will not work with bytes in python-3.1:
                    #   TypeError: can't concat bytes to str
                    if gdbm is None:
                        self.__db = dbm.open(self._db_path, "c", self._perms)
                    else:
                        # Prefer gdbm type if available, since it allows
                        # multiple concurrent writers (see bug #53607).
                        self.__db = gdbm.open(self._db_path, "cu", self._perms)
            except dbm.error as e:
                raise cache_errors.InitializationError(self.__class__, e)
        self._ensure_access(self._db_path)

    def iteritems(self):
        # dbm doesn't implement items()
        for k in self.__db.keys():
            yield (k, self[k])

    def _getitem(self, cpv):
        # we override getitem because it's just a cpickling of the data handed in.
        return pickle.loads(self.__db[_unicode_encode(cpv)])

    def _setitem(self, cpv, values):
        self.__db[_unicode_encode(cpv)] = pickle.dumps(values, pickle.HIGHEST_PROTOCOL)

    def _delitem(self, cpv):
        del self.__db[cpv]

    def __iter__(self):
        return iter(list(self.__db.keys()))

    def __contains__(self, cpv):
        return cpv in self.__db

    def __del__(self):
        if "__db" in self.__dict__ and self.__db is not None:
            self.__db.sync()
            self.__db.close()

    # TODO: do we need iteritems()?
    items = iteritems



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import subprocess
import tempfile

from portage.const import PORTAGE_BIN_PATH
from portage.tests import TestCase


class TestEAPI7VerFuncs(TestCase):
    def _test_output(self, test_cases):
        """
        Test that commands in test_cases produce expected output.
        """
        with tempfile.NamedTemporaryFile("w") as test_script:
            test_script.write('source "%s"/eapi7-ver-funcs.sh\n' % (PORTAGE_BIN_PATH,))
            for cmd, exp in test_cases:
                test_script.write("%s\n" % (cmd,))
            test_script.flush()

            s = subprocess.Popen(
                ["bash", test_script.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            sout, serr = s.communicate()
            self.assertEqual(s.returncode, 0)

            for test_case, result in zip(test_cases, sout.decode().splitlines()):
                cmd, exp = test_case
                self.assertEqual(
                    result, exp, "%s -> %s; expected: %s" % (cmd, result, exp)
                )

    def _test_return(self, test_cases):
        """
        Test that commands in test_cases give appropriate exit codes.
        """
        with tempfile.NamedTemporaryFile("w+") as test_script:
            test_script.write('source "%s"/eapi7-ver-funcs.sh\n' % (PORTAGE_BIN_PATH,))
            for cmd, exp in test_cases:
                test_script.write("%s; echo $?\n" % (cmd,))
            test_script.flush()

            s = subprocess.Popen(
                ["bash", test_script.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            sout, serr = s.communicate()
            self.assertEqual(s.returncode, 0)

            for test_case, result in zip(test_cases, sout.decode().splitlines()):
                cmd, exp = test_case
                self.assertEqual(
                    result, exp, "%s -> %s; expected: %s" % (cmd, result, exp)
                )

    def _test_fail(self, test_cases):
        """
        Test that commands in test_cases fail.
        """

        for cmd in test_cases:
            test = """
source "%s"/eapi7-ver-funcs.sh
die() { exit 1; }
%s""" % (
                PORTAGE_BIN_PATH,
                cmd,
            )

            s = subprocess.Popen(
                ["bash", "-c", test], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            sout, serr = s.communicate()
            self.assertEqual(
                s.returncode,
                1,
                '"%s" did not fail; output: %s; %s)'
                % (cmd, sout.decode(), serr.decode()),
            )

    def test_ver_cut(self):
        test_cases = [
            # (command, output)
            ("ver_cut 1 1.2.3", "1"),
            ("ver_cut 1-1 1.2.3", "1"),
            ("ver_cut 1-2 1.2.3", "1.2"),
            ("ver_cut 2- 1.2.3", "2.3"),
            ("ver_cut 1- 1.2.3", "1.2.3"),
            ("ver_cut 3-4 1.2.3b_alpha4", "3b"),
            ("ver_cut 5 1.2.3b_alpha4", "alpha"),
            ("ver_cut 1-2 .1.2.3", "1.2"),
            ("ver_cut 0-2 .1.2.3", ".1.2"),
            ("ver_cut 2-3 1.2.3.", "2.3"),
            ("ver_cut 2- 1.2.3.", "2.3."),
            ("ver_cut 2-4 1.2.3.", "2.3."),
        ]
        self._test_output(test_cases)

    def test_ver_rs(self):
        test_cases = [
            # (command, output)
            ("ver_rs 1 - 1.2.3", "1-2.3"),
            ("ver_rs 2 - 1.2.3", "1.2-3"),
            ("ver_rs 1-2 - 1.2.3.4", "1-2-3.4"),
            ("ver_rs 2- - 1.2.3.4", "1.2-3-4"),
            ("ver_rs 2 . 1.2-3", "1.2.3"),
            ("ver_rs 3 . 1.2.3a", "1.2.3.a"),
            ("ver_rs 2-3 - 1.2_alpha4", "1.2-alpha-4"),
            ('ver_rs 3 - 2 "" 1.2.3b_alpha4', "1.23-b_alpha4"),
            ("ver_rs 3-5 _ 4-6 - a1b2c3d4e5", "a1b_2-c-3-d4e5"),
            ("ver_rs 1 - .1.2.3", ".1-2.3"),
            ("ver_rs 0 - .1.2.3", "-1.2.3"),
        ]
        self._test_output(test_cases)

    def test_truncated_range(self):
        test_cases = [
            # (command, output)
            ("ver_cut 0-2 1.2.3", "1.2"),
            ("ver_cut 2-5 1.2.3", "2.3"),
            ("ver_cut 4 1.2.3", ""),
            ("ver_cut 0 1.2.3", ""),
            ("ver_cut 4- 1.2.3", ""),
            ("ver_rs 0 - 1.2.3", "1.2.3"),
            ("ver_rs 3 . 1.2.3", "1.2.3"),
            ("ver_rs 3- . 1.2.3", "1.2.3"),
            ("ver_rs 3-5 . 1.2.3", "1.2.3"),
        ]
        self._test_output(test_cases)

    def test_invalid_range(self):
        test_cases = [
            "ver_cut foo 1.2.3",
            "ver_rs -3 _ a1b2c3d4e5",
            "ver_rs 5-3 _ a1b2c3d4e5",
        ]
        self._test_fail(test_cases)

    def test_ver_test(self):
        test_cases = [
            # Tests from Portage's test_vercmp.py
            ("ver_test 6.0 -gt 5.0", "0"),
            ("ver_test 5.0 -gt 5", "0"),
            ("ver_test 1.0-r1 -gt 1.0-r0", "0"),
            ("ver_test 999999999999999999 -gt 999999999999999998", "0"),  # 18 digits
            ("ver_test 1.0.0 -gt 1.0", "0"),
            ("ver_test 1.0.0 -gt 1.0b", "0"),
            ("ver_test 1b -gt 1", "0"),
            ("ver_test 1b_p1 -gt 1_p1", "0"),
            ("ver_test 1.1b -gt 1.1", "0"),
            ("ver_test 12.2.5 -gt 12.2b", "0"),
            ("ver_test 4.0 -lt 5.0", "0"),
            ("ver_test 5 -lt 5.0", "0"),
            ("ver_test 1.0_pre2 -lt 1.0_p2", "0"),
            ("ver_test 1.0_alpha2 -lt 1.0_p2", "0"),
            ("ver_test 1.0_alpha1 -lt 1.0_beta1", "0"),
            ("ver_test 1.0_beta3 -lt 1.0_rc3", "0"),
            ("ver_test 1.001000000000000001 -lt 1.001000000000000002", "0"),
            ("ver_test 1.00100000000 -lt 1.001000000000000001", "0"),
            ("ver_test 999999999999999998 -lt 999999999999999999", "0"),
            ("ver_test 1.01 -lt 1.1", "0"),
            ("ver_test 1.0-r0 -lt 1.0-r1", "0"),
            ("ver_test 1.0 -lt 1.0-r1", "0"),
            ("ver_test 1.0 -lt 1.0.0", "0"),
            ("ver_test 1.0b -lt 1.0.0", "0"),
            ("ver_test 1_p1 -lt 1b_p1", "0"),
            ("ver_test 1 -lt 1b", "0"),
            ("ver_test 1.1 -lt 1.1b", "0"),
            ("ver_test 12.2b -lt 12.2.5", "0"),
            ("ver_test 4.0 -eq 4.0", "0"),
            ("ver_test 1.0 -eq 1.0", "0"),
            ("ver_test 1.0-r0 -eq 1.0", "0"),
            ("ver_test 1.0 -eq 1.0-r0", "0"),
            ("ver_test 1.0-r0 -eq 1.0-r0", "0"),
            ("ver_test 1.0-r1 -eq 1.0-r1", "0"),
            ("ver_test 1 -eq 2", "1"),
            ("ver_test 1.0_alpha -eq 1.0_pre", "1"),
            ("ver_test 1.0_beta -eq 1.0_alpha", "1"),
            ("ver_test 1 -eq 0.0", "1"),
            ("ver_test 1.0-r0 -eq 1.0-r1", "1"),
            ("ver_test 1.0-r1 -eq 1.0-r0", "1"),
            ("ver_test 1.0 -eq 1.0-r1", "1"),
            ("ver_test 1.0-r1 -eq 1.0", "1"),
            ("ver_test 1.0 -eq 1.0.0", "1"),
            ("ver_test 1_p1 -eq 1b_p1", "1"),
            ("ver_test 1b -eq 1", "1"),
            ("ver_test 1.1b -eq 1.1", "1"),
            ("ver_test 12.2b -eq 12.2", "1"),
            # A subset of tests from Paludis
            ("ver_test 1.0_alpha -gt 1_alpha", "0"),
            ("ver_test 1.0_alpha -gt 1", "0"),
            ("ver_test 1.0_alpha -lt 1.0", "0"),
            ("ver_test 1.2.0.0_alpha7-r4 -gt 1.2_alpha7-r4", "0"),
            ("ver_test 0001 -eq 1", "0"),
            ("ver_test 01 -eq 001", "0"),
            ("ver_test 0001.1 -eq 1.1", "0"),
            ("ver_test 01.01 -eq 1.01", "0"),
            ("ver_test 1.010 -eq 1.01", "0"),
            ("ver_test 1.00 -eq 1.0", "0"),
            ("ver_test 1.0100 -eq 1.010", "0"),
            ("ver_test 1-r00 -eq 1-r0", "0"),
            # Additional tests
            ("ver_test 0_rc99 -lt 0", "0"),
            ("ver_test 011 -eq 11", "0"),
            ("ver_test 019 -eq 19", "0"),
            ("ver_test 1.2 -eq 001.2", "0"),
            ("ver_test 1.2 -gt 1.02", "0"),
            ("ver_test 1.2a -lt 1.2b", "0"),
            ("ver_test 1.2_pre1 -gt 1.2_pre1_beta2", "0"),
            ("ver_test 1.2_pre1 -lt 1.2_pre1_p2", "0"),
            ("ver_test 1.00 -lt 1.0.0", "0"),
            ("ver_test 1.010 -eq 1.01", "0"),
            ("ver_test 1.01 -lt 1.1", "0"),
            ("ver_test 1.2_pre08-r09 -eq 1.2_pre8-r9", "0"),
            ("ver_test 0 -lt 576460752303423488", "0"),  # 2**59
            ("ver_test 0 -lt 9223372036854775808", "0"),  # 2**63
        ]
        self._test_return(test_cases)

    def test_invalid_test(self):
        test_cases = [
            # Bad number or ordering of arguments
            "ver_test 1",
            "ver_test 1 -lt 2 3",
            "ver_test -lt 1 2",
            # Bad operators
            'ver_test 1 "<" 2',
            "ver_test 1 lt 2",
            "ver_test 1 -foo 2",
            # Malformed versions
            'ver_test "" -ne 1',
            "ver_test 1. -ne 1",
            "ver_test 1ab -ne 1",
            "ver_test b -ne 1",
            "ver_test 1-r1_pre -ne 1",
            "ver_test 1-pre1 -ne 1",
            "ver_test 1_foo -ne 1",
            "ver_test 1_pre1.1 -ne 1",
            "ver_test 1-r1.0 -ne 1",
            "ver_test cvs.9999 -ne 9999",
        ]
        self._test_fail(test_cases)



#!/usr/bin/env python
# Copyright 2009-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import argparse
import sys
import portage

portage._internal_caller = True
from portage import os


def command_recompose(args):

    usage = "usage: recompose <binpkg_path> <metadata_dir>\n"

    if len(args) != 2:
        sys.stderr.write(usage)
        sys.stderr.write("2 arguments are required, got %s\n" % len(args))
        return 1

    binpkg_path, metadata_dir = args

    if not os.path.isfile(binpkg_path):
        sys.stderr.write(usage)
        sys.stderr.write("Argument 1 is not a regular file: '%s'\n" % binpkg_path)
        return 1

    if not os.path.isdir(metadata_dir):
        sys.stderr.write(usage)
        sys.stderr.write("Argument 2 is not a directory: '%s'\n" % metadata_dir)
        return 1

    t = portage.xpak.tbz2(binpkg_path)
    t.recompose(metadata_dir)
    return os.EX_OK


def main(argv):

    if argv and isinstance(argv[0], bytes):
        for i, x in enumerate(argv):
            argv[i] = portage._unicode_decode(x, errors="strict")

    valid_commands = ("recompose",)
    description = "Perform metadata operations on a binary package."
    usage = "usage: %s COMMAND [args]" % os.path.basename(argv[0])

    parser = argparse.ArgumentParser(description=description, usage=usage)
    options, args = parser.parse_known_args(argv[1:])

    if not args:
        parser.error("missing command argument")

    command = args[0]

    if command not in valid_commands:
        parser.error("invalid command: '%s'" % command)

    if command == "recompose":
        rval = command_recompose(args[1:])
    else:
        raise AssertionError("invalid command: '%s'" % command)

    return rval


if __name__ == "__main__":
    rval = main(sys.argv[:])
    sys.exit(rval)



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class MultSlotTestCase(TestCase):
    def testMultiSlotSelective(self):
        """
        Test that a package isn't reinstalled due to SLOT dependency
        interaction with USE=multislot (bug #220341).
        """

        ebuilds = {
            "sys-devel/gcc-4.4.4": {"SLOT": "4.4"},
            "dev-util/nvidia-cuda-toolkit-4.0": {
                "EAPI": "1",
                "RDEPEND": "sys-devel/gcc:4.4",
            },
        }

        installed = {
            "sys-devel/gcc-4.4.4": {"SLOT": "i686-pc-linux-gnu-4.4.4"},
            "dev-util/nvidia-cuda-toolkit-4.0": {
                "EAPI": "1",
                "RDEPEND": "sys-devel/gcc:4.4",
            },
        }

        world = ("dev-util/nvidia-cuda-toolkit",)

        options = {"--update": True, "--deep": True, "--selective": True}

        test_cases = (
            ResolverPlaygroundTestCase(
                ["sys-devel/gcc:4.4"], options=options, mergelist=[], success=True
            ),
            # depclean test for bug #382823
            ResolverPlaygroundTestCase(
                [], options={"--depclean": True}, success=True, cleanlist=[]
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class ImageMagickGraphicsMagickTestCase(TestCase):
    def testImageMagickUpdate(self):

        ebuilds = {
            "media-gfx/imagemagick-6.9.7.0": {
                "EAPI": "6",
                "SLOT": "0/6.9.7.0",
            },
            "media-gfx/imagemagick-6.9.6.6": {
                "EAPI": "6",
                "SLOT": "0/6.9.6.6",
            },
            "media-gfx/inkscape-0.91-r3": {
                "EAPI": "6",
                "DEPEND": "media-gfx/imagemagick:=",
                "RDEPEND": "media-gfx/imagemagick:=",
            },
            "media-video/dvdrip-0.98.11-r3": {
                "EAPI": "6",
                "DEPEND": "|| ( media-gfx/graphicsmagick[imagemagick] media-gfx/imagemagick )",
                "RDEPEND": "|| ( media-gfx/graphicsmagick[imagemagick] media-gfx/imagemagick )",
            },
            "media-gfx/graphicsmagick-1.3.25": {
                "EAPI": "6",
                "SLOT": "0/1.3",
                "IUSE": "imagemagick",
                "RDEPEND": "imagemagick? ( !media-gfx/imagemagick )",
            },
        }

        installed = {
            "media-gfx/imagemagick-6.9.6.6": {
                "EAPI": "6",
                "SLOT": "0/6.9.6.6",
            },
            "media-gfx/inkscape-0.91-r3": {
                "EAPI": "6",
                "DEPEND": "media-gfx/imagemagick:0/6.9.6.6=",
                "RDEPEND": "media-gfx/imagemagick:0/6.9.6.6=",
            },
            "media-video/dvdrip-0.98.11-r3": {
                "EAPI": "6",
                "DEPEND": "|| ( media-gfx/graphicsmagick[imagemagick] media-gfx/imagemagick )",
                "RDEPEND": "|| ( media-gfx/graphicsmagick[imagemagick] media-gfx/imagemagick )",
            },
            "media-gfx/graphicsmagick-1.3.25": {
                "EAPI": "6",
                "SLOT": "0/1.3",
                "IUSE": "imagemagick",
                "USE": "",
                "RDEPEND": "imagemagick? ( !media-gfx/imagemagick )",
            },
        }

        world = (
            "media-gfx/inkscape",
            "media-video/dvdrip",
            "media-gfx/graphicsmagick",
        )

        test_cases = (
            # bug #554070: imagemagick upgrade triggered erroneous
            # autounmask USE change for media-gfx/graphicsmagick[imagemagick]
            ResolverPlaygroundTestCase(
                ["media-gfx/imagemagick", "@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "media-gfx/imagemagick-6.9.7.0",
                    "media-gfx/inkscape-0.91-r3",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            # Disable debug so that cleanup works.
            playground.debug = False
            playground.cleanup()



# Copyright 2018-2022 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import tempfile
from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.util.install_mask import InstallMask, install_mask_dir


class InstallMaskTestCase(TestCase):
    def testTrailingSlash(self):
        """
        Test that elements with a trailing slash match a directory
        but not a regular file.
        """
        cases = (
            (
                "/foo/bar/ -/foo/bar/*.foo -*.baz",
                (
                    (
                        "foo/bar/baz",
                        True,
                    ),
                    (
                        "foo/bar/",
                        True,
                    ),
                    # /foo/bar/ does not match
                    (
                        "foo/bar",
                        False,
                    ),
                    # this is excluded
                    (
                        "foo/bar/baz.foo",
                        False,
                    ),
                    # this is excluded
                    (
                        "foo/bar/baz.baz",
                        False,
                    ),
                    (
                        "foo/bar/baz.bar",
                        True,
                    ),
                ),
            ),
            (
                "/foo/bar -/foo/bar/*.foo -*.baz",
                (
                    (
                        "foo/bar/baz",
                        True,
                    ),
                    # /foo/bar matches both foo/bar/ and foo/bar
                    (
                        "foo/bar/",
                        True,
                    ),
                    (
                        "foo/bar",
                        True,
                    ),
                    # this is excluded
                    (
                        "foo/bar/baz.foo",
                        False,
                    ),
                    # this is excluded
                    (
                        "foo/bar/baz.baz",
                        False,
                    ),
                    (
                        "foo/bar/baz.bar",
                        True,
                    ),
                ),
            ),
            (
                "/foo*",
                (
                    (
                        "foo",
                        True,
                    ),
                    (
                        "foo/",
                        True,
                    ),
                    (
                        "foobar",
                        True,
                    ),
                    (
                        "foobar/",
                        True,
                    ),
                ),
            ),
            (
                "/foo*/",
                (
                    (
                        "foo",
                        False,
                    ),
                    (
                        "foo/",
                        True,
                    ),
                    (
                        "foobar",
                        False,
                    ),
                    (
                        "foobar/",
                        True,
                    ),
                ),
            ),
            (
                "/usr/share/locale "
                "-/usr/share/locale/en* "
                "-/usr/share/locale/kf5_all_languages "
                "-/usr/share/locale/locale.alias",
                (
                    (
                        "usr/share/locale/en",
                        False,
                    ),
                    (
                        "usr/share/locale/en_GB",
                        False,
                    ),
                    (
                        "usr/share/locale/en/kf5_all_languages",
                        False,
                    ),
                    (
                        "usr/share/locale/locale.alias",
                        False,
                    ),
                    (
                        "usr/share/locale/es",
                        True,
                    ),
                    (
                        "usr/share/locale/fr",
                        True,
                    ),
                    (
                        "usr/share/locale",
                        True,
                    ),
                ),
            ),
        )

        for install_mask_str, paths in cases:
            install_mask = InstallMask(install_mask_str)
            for path, expected in paths:
                self.assertEqual(
                    install_mask.match(path),
                    expected,
                    'unexpected match result for "{}" with path {}'.format(
                        install_mask_str, path
                    ),
                )

    def testSymlinkDir(self):
        """
        Test that masked symlinks to directories are removed.
        """
        tmp_dir = tempfile.mkdtemp()

        try:
            base_dir = os.path.join(tmp_dir, "foo")
            target_dir = os.path.join(tmp_dir, "foo", "bar")
            link_name = os.path.join(tmp_dir, "foo", "baz")

            os.mkdir(base_dir)
            os.mkdir(target_dir)
            os.symlink(target_dir, link_name)

            install_mask = InstallMask("/foo/")
            install_mask_dir(tmp_dir, install_mask)
            self.assertFalse(
                os.path.lexists(link_name), "failed to remove {}".format(link_name)
            )
            self.assertFalse(
                os.path.lexists(base_dir), "failed to remove {}".format(base_dir)
            )
        finally:
            shutil.rmtree(tmp_dir)



# Copyright 2008-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


__all__ = ["ObjectProxy"]


class ObjectProxy:

    """
    Object that acts as a proxy to another object, forwarding
    attribute accesses and method calls. This can be useful
    for implementing lazy initialization.
    """

    __slots__ = ()

    def _get_target(self):
        raise NotImplementedError(self)

    def __getattribute__(self, attr):
        result = object.__getattribute__(self, "_get_target")()
        return getattr(result, attr)

    def __setattr__(self, attr, value):
        result = object.__getattribute__(self, "_get_target")()
        setattr(result, attr, value)

    def __call__(self, *args, **kwargs):
        result = object.__getattribute__(self, "_get_target")()
        return result(*args, **kwargs)

    def __enter__(self):
        return object.__getattribute__(self, "_get_target")().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        return object.__getattribute__(self, "_get_target")().__exit__(
            exc_type, exc_value, traceback
        )

    def __setitem__(self, key, value):
        object.__getattribute__(self, "_get_target")()[key] = value

    def __getitem__(self, key):
        return object.__getattribute__(self, "_get_target")()[key]

    def __delitem__(self, key):
        del object.__getattribute__(self, "_get_target")()[key]

    def __contains__(self, key):
        return key in object.__getattribute__(self, "_get_target")()

    def __iter__(self):
        return iter(object.__getattribute__(self, "_get_target")())

    def __len__(self):
        return len(object.__getattribute__(self, "_get_target")())

    def __repr__(self):
        return repr(object.__getattribute__(self, "_get_target")())

    def __str__(self):
        return str(object.__getattribute__(self, "_get_target")())

    def __add__(self, other):
        return self.__str__() + other

    def __hash__(self):
        return hash(object.__getattribute__(self, "_get_target")())

    def __ge__(self, other):
        return object.__getattribute__(self, "_get_target")() >= other

    def __gt__(self, other):
        return object.__getattribute__(self, "_get_target")() > other

    def __le__(self, other):
        return object.__getattribute__(self, "_get_target")() <= other

    def __lt__(self, other):
        return object.__getattribute__(self, "_get_target")() < other

    def __eq__(self, other):
        return object.__getattribute__(self, "_get_target")() == other

    def __ne__(self, other):
        return object.__getattribute__(self, "_get_target")() != other

    def __bool__(self):
        return bool(object.__getattribute__(self, "_get_target")())

    def __int__(self):
        return int(object.__getattribute__(self, "_get_target")())



# Copright Gentoo Foundation 2006-2020
# Portage Unit Testing Functionality

import io
import tarfile
import tempfile
from os import urandom

from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.gpkg import gpkg
from portage.gpg import GPG
from portage.exception import MissingSignature, InvalidSignature


class test_gpkg_gpg_case(TestCase):
    def test_gpkg_missing_manifest_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                    'BINPKG_FORMAT="gpkg"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == os.path.join("test", "Manifest"):
                            manifest = tar_1.extractfile(f).read().decode("UTF-8")
                            manifest = manifest.replace(
                                "-----BEGIN PGP SIGNATURE-----", ""
                            )
                            manifest = manifest.replace(
                                "-----END PGP SIGNATURE-----", ""
                            )
                            manifest_data = io.BytesIO(manifest.encode("UTF-8"))
                            manifest_data.seek(0, io.SEEK_END)
                            f.size = manifest_data.tell()
                            manifest_data.seek(0)
                            tar_2.addfile(f, manifest_data)
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))

            self.assertRaises(
                InvalidSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_missing_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                    'BINPKG_FORMAT="gpkg"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name.endswith(".sig"):
                            pass
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))
            self.assertRaises(
                MissingSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )

        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_ignore_signature(self):
        gpg_test_path = os.environ["PORTAGE_GNUPGHOME"]

        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-ignore-signature"',
                    'BINPKG_FORMAT="gpkg"',
                    f'BINPKG_GPG_SIGNING_BASE_COMMAND="flock {gpg_test_path}/portage-binpkg-gpg.lock /usr/bin/gpg --sign --armor --batch --no-tty --yes --pinentry-mode loopback --passphrase GentooTest [PORTAGE_CONFIG]"',
                    'BINPKG_GPG_SIGNING_DIGEST="SHA512"',
                    f'BINPKG_GPG_SIGNING_GPG_HOME="{gpg_test_path}"',
                    'BINPKG_GPG_SIGNING_KEY="0x8812797DDF1DD192"',
                    'BINPKG_GPG_VERIFY_BASE_COMMAND="/usr/bin/gpg --verify --batch --no-tty --yes --no-auto-check-trustdb --status-fd 2 [PORTAGE_CONFIG] [SIGNATURE]"',
                    f'BINPKG_GPG_VERIFY_GPG_HOME="{gpg_test_path}"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_2.decompress(os.path.join(tmpdir, "test"))
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_auto_use_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing '
                    '-binpkg-request-signature"',
                    'BINPKG_FORMAT="gpkg"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name.endswith(".sig"):
                            pass
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))
            self.assertRaises(
                MissingSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_invalid_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                    'BINPKG_FORMAT="gpkg"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == os.path.join("test", "Manifest"):
                            sig = b"""
-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA512

DATA test/image.tar.zst 1049649 BLAKE2B 3112adba9c09023962f26d9dcbf8e74107c05220f2f29aa2ce894f8a4104c3bb238f87095df73735befcf1e1f6039fc3abf4defa87e68ce80f33dd01e09c055a SHA512 9f584727f2e20a50a30e0077b94082c8c1f517ebfc9978eb3281887e24458108e73d1a2ce82eb0b59f5df7181597e4b0a297ae68bbfb36763aa052e6bdbf2c59
DATA test/image.tar.zst.sig 833 BLAKE2B 214724ae4ff9198879c8c960fd8167632e27982c2278bb873f195abe75b75afa1ebed4c37ec696f5f5bc35c3a1184b60e0b50d56695b072b254f730db01eddb5 SHA512 67316187da8bb6b7a5f9dc6a42ed5c7d72c6184483a97f23c0bebd8b187ac9268e0409eb233c935101606768718c99eaa5699037d6a68c2d88c9ed5331a3f73c
-----BEGIN PGP SIGNATURE-----

iQIzBAEBCgAdFiEEBrOjEb13XCgNIqkwXZDqBjUhd/YFAmFazXEACgkQXZDqBjUh
d/YFZA//eiXkYAS2NKxim6Ppr1HcZdjU1f6H+zyQzC7OdPkAh7wsVXpSr1aq+giD
G4tNtI6nsFokpA5CMhDf+ffBofKmFY5plk9zyQHr43N/RS5G6pcb2LHk0mQqgIdB
EsZRRD75Na4uGDWjuNHRmsasPTsc9qyW7FLckjwUsVmk9foAoiLYYaTsilsEGqXD
Bl/Z6PaQXvdd8txbcP6dOXfhVT06b+RWcnHI06KQrmFkZjZQh/7bCIeCVwNbXr7d
Obo8SVzCrQbTONei57AkyuRfnPqBfP61k8rQtcDUmCckQQfyaRwoW2nDIewOPfIH
xfvM137to2GEI2RR1TpWmGfu3iQzgC71f4svdX9Tyi5N7aFmfud7LZs6/Un3IdVk
ZH9/AmRzeH6hKllqSv/6WuhjsTNvr0bOzGbskkhqlLga2tml08gHFYOMWRJb/bRz
N8FZMhHzFoc0hsG8SU9uC+OeW+y5NdqpbRnQwgABmAiKEpgAPnABTsr0HjyxvjY+
uCUdvMMHvnTxTjNEZ3Q+UQ2VsSoZzPbW9Y4PuM0XxxmTI8htdn4uIhy9dLNPsJmB
eTE8aov/1uKq9VMsYC8wcx5vLMaR7/O/9XstP+r6PaZwiLlyrKHGexV4O52sj6LC
qGAN3VUF+8EsdcsV781H0F86PANhyBgEYTGDrnItTGe3/vAPjCo=
=S/Vn
-----END PGP SIGNATURE-----
"""
                            data = io.BytesIO(sig)
                            f.size = len(sig)
                            tar_2.addfile(f, data)
                            data.close()
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))
            self.assertRaises(
                InvalidSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_untrusted_signature(self):
        gpg_test_path = os.environ["PORTAGE_GNUPGHOME"]

        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                    'BINPKG_FORMAT="gpkg"',
                    f'BINPKG_GPG_SIGNING_BASE_COMMAND="flock {gpg_test_path}/portage-binpkg-gpg.lock /usr/bin/gpg --sign --armor --batch --no-tty --yes --pinentry-mode loopback --passphrase GentooTest [PORTAGE_CONFIG]"',
                    'BINPKG_GPG_SIGNING_DIGEST="SHA512"',
                    f'BINPKG_GPG_SIGNING_GPG_HOME="{gpg_test_path}"',
                    'BINPKG_GPG_SIGNING_KEY="0x8812797DDF1DD192"',
                    'BINPKG_GPG_VERIFY_BASE_COMMAND="/usr/bin/gpg --verify --batch --no-tty --yes --no-auto-check-trustdb --status-fd 2 [PORTAGE_CONFIG] [SIGNATURE]"',
                    f'BINPKG_GPG_VERIFY_GPG_HOME="{gpg_test_path}"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            self.assertRaises(
                InvalidSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )

        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_unknown_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                    'BINPKG_FORMAT="gpkg"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            data = urandom(1048576)
            with open(os.path.join(orig_full_path, "data"), "wb") as f:
                f.write(data)

            binpkg_1 = gpkg(settings, "test", os.path.join(tmpdir, "test-1.gpkg.tar"))
            binpkg_1.compress(orig_full_path, {})

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == os.path.join("test", "Manifest"):
                            sig = b"""
-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256


DATA test/image.tar.zst 1049649 BLAKE2B 3112adba9c09023962f26d9dcbf8e74107c05220f2f29aa2ce894f8a4104c3bb238f87095df73735befcf1e1f6039fc3abf4defa87e68ce80f33dd01e09c055a SHA512 9f584727f2e20a50a30e0077b94082c8c1f517ebfc9978eb3281887e24458108e73d1a2ce82eb0b59f5df7181597e4b0a297ae68bbfb36763aa052e6bdbf2c59
DATA test/image.tar.zst.sig 833 BLAKE2B 214724ae4ff9198879c8c960fd8167632e27982c2278bb873f195abe75b75afa1ebed4c37ec696f5f5bc35c3a1184b60e0b50d56695b072b254f730db01eddb5 SHA512 67316187da8bb6b7a5f9dc6a42ed5c7d72c6184483a97f23c0bebd8b187ac9268e0409eb233c935101606768718c99eaa5699037d6a68c2d88c9ed5331a3f73c
-----BEGIN PGP SIGNATURE-----

iNUEARYIAH0WIQSMe+CQzU+/D/DeMitA3PGOlxUHlQUCYVrQal8UgAAAAAAuAChp
c3N1ZXItZnByQG5vdGF0aW9ucy5vcGVucGdwLmZpZnRoaG9yc2VtYW4ubmV0OEM3
QkUwOTBDRDRGQkYwRkYwREUzMjJCNDBEQ0YxOEU5NzE1MDc5NQAKCRBA3PGOlxUH
lbmTAP4jdhMTW6g550/t0V7XcixqVtBockOTln8hZrZIQrjAJAD/caDkxgz5Xl8C
EP1pgSXXGtlUnv6akg/wueFJKEr9KQs=
=edEg
-----END PGP SIGNATURE-----
"""
                            data = io.BytesIO(sig)
                            f.size = len(sig)
                            tar_2.addfile(f, data)
                            data.close()
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            binpkg_2 = gpkg(settings, "test", os.path.join(tmpdir, "test-2.gpkg.tar"))
            self.assertRaises(
                InvalidSignature, binpkg_2.decompress, os.path.join(tmpdir, "test")
            )

        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()
